import copy

from qiskit import QuantumCircuit, transpile
from qiskit_ibm_runtime import QiskitRuntimeService

from .base import QuantumBackend


class IBMBackend(QuantumBackend):
    """
    Backend using IQM device for circuit execution
    """

    def __init__(
        self,
        instance: str | None = None,
        device: str | None = None,
        min_num_qubits: int | None = None,
    ) -> None:
        """Constructor

        Args:
            instance: Access instance for premium devices, otherwise use free devices
            device: Device name to use, defaults to least busy
            min_num_qubits: If getting least busy device, provide the minimum number of qubits required
        """
        self.instance = instance
        self.service = QiskitRuntimeService(channel=None, instance=self.instance)
        if device is None:
            self.backend = self.service.least_busy(
                min_num_qubits=min_num_qubits, instance=self.instance
            )
        else:
            self.backend = self.service.backend(name=device)
        self._name = self.backend.name
        self._coupling_map = self.backend.coupling_map
        self._basis_gates = self.backend.operation_names
        self._num_qubits = self.backend.num_qubits

    @property
    def name(self) -> str:
        return self._name

    @property
    def coupling_map(self) -> list[tuple[int, int]]:
        return self._coupling_map

    @property
    def num_qubits(self) -> int:
        return self._num_qubits

    @property
    def basis_gates(self) -> list[str]:
        return self._basis_gates

    def run(
        self,
        circuit: QuantumCircuit,
        shots: int,
    ) -> dict[str, int]:
        """Execute the circuit

        Args:
            circuit: The QuantumCircuit object to run
            shots: If provided will sample from the resulting state
            max_bond: The maximum bond dimension allowed

        Returns:
            Measurement results {bitstring : count}
        """
        qc = copy.deepcopy(circuit)
        qc.measure_all()
        qc = transpile(qc, backend=self.backend)
        result = self.backend.run(qc, shots=shots).result()
        counts = result.get_counts()
        counts = {k[::-1]: v for k, v in counts.items()}
        return counts

    def parse_openqasm(self, filename: str) -> QuantumCircuit:
        """Parse an OpenQASM input circuit

        Args:
            filename: The filename of the OpenQASM input

        Returns:
            A qiskit QuantumCircuit object
        """
        qc = QuantumCircuit.from_qasm_file(filename)
        qc = transpile(qc, backend=self.backend)
        return qc

    def get_device_info(self) -> dict:
        """Return a dictionary describing the backend."""
        return {"backend_name": self.name}



import copy

import numpy as np

from ..mpo import MatrixProductOperator
from ..mps import MatrixProductState


class HFSuppression:
    def __init__(
        self,
        hf_state: str,
        mps: MatrixProductState,
        mpo: MatrixProductOperator | None = None,
    ):
        """Constructor

        Args:
            hf_state: The HF state of the system
            mps: The state to suppress the HF contribution
            mpo: The operator to evolve
        """
        self.hf_state = hf_state
        self.num_electrons = hf_state.count("1")
        self.num_qubits = mps.num_sites
        self.hf_state_mps = MatrixProductState.from_bitstring(hf_state)
        self.mps = mps
        self.mpo = mpo
        self.suppressed_mps = None
        self.evolved_mpo = None

    def build_hf_oracle(self) -> MatrixProductOperator:
        """Build oracle for HF amplitude suppression"""
        oracle_mpo = self.hf_state_mps.form_density_operator()
        oracle_mpo.multiply_by_constant(2.0)
        id_mpo = MatrixProductOperator.identity_mpo(oracle_mpo.num_sites)
        oracle_mpo = oracle_mpo - id_mpo

        return oracle_mpo

    def build_diffusion_operator(self) -> MatrixProductOperator:
        """Build diffusion operator circuit"""
        diffusion_mpo = self.mps.form_density_operator()
        diffusion_mpo.multiply_by_constant(2.0)
        id_mpo = MatrixProductOperator.identity_mpo(self.mps.num_sites)
        diffusion_mpo = diffusion_mpo - id_mpo

        return diffusion_mpo

    def hf_suppression(self, max_bond: int | None = None) -> MatrixProductState:
        """Perform HF suppression on MPS

        Args:
            max_bond: Maximum bond dimension

        Returns:
            The HF suppressed MPS
        """
        oracle = self.build_hf_oracle()
        diffusion = self.build_diffusion_operator()
        hf_amplitude = self.mps.compute_inner_product(self.hf_state_mps)
        rotation_angle = np.sqrt(1 - np.abs(hf_amplitude) ** 2)
        num_iterations = int(np.floor(np.pi / (4 * rotation_angle)))
        current_mps = copy.deepcopy(self.mps)
        for _ in range(num_iterations):
            current_mps = current_mps.apply_mpo(oracle, max_bond)
            current_mps = current_mps.apply_mpo(diffusion, max_bond)
        return current_mps

    def operator_evolution(self, max_bond: int | None = None) -> MatrixProductOperator:
        """Evolve the MPO in line with HF suppression

        Args:
            max_bond: Maximum bond dimension

        Returns:
            The evolved MPO
        """
        oracle = self.build_hf_oracle()
        oracle_dag = copy.deepcopy(oracle)
        oracle_dag.dagger()
        diffusion = self.build_diffusion_operator()
        diffusion_dag = copy.deepcopy(diffusion)
        diffusion_dag.dagger()
        hf_amplitude = self.mps.compute_inner_product(self.hf_state_mps)
        rotation_angle = np.sqrt(1 - np.abs(hf_amplitude) ** 2)
        num_iterations = int(np.floor(np.pi / (4 * rotation_angle)))
        current_mpo = copy.deepcopy(self.mpo)
        for _ in range(num_iterations):
            current_mpo = oracle_dag * current_mpo
            current_mpo = diffusion_dag * current_mpo
            current_mpo = current_mpo * oracle
            current_mpo = current_mpo * diffusion
            if max_bond:
                if current_mpo.bond_dimension > max_bond:
                    current_mpo.compress(max_bond)
        return current_mpo

    def run(self, max_bond: int | None = None) -> None:
        """Run HF Suppressio"""
        self.suppressed_mps = self.hf_suppression(max_bond)
        self.evolved_mpo = self.operator_evolution(max_bond)
        return



from typing import Callable

import numpy as np
from numpy import ndarray
from openfermion.ops import FermionOperator
from openfermion.transforms import jordan_wigner
from scipy.optimize import minimize
from symmer.operators import PauliwordOp

from tn4qa.qi_cost_functions import (
    cost_function_dict_to_purity_mpo,
    cost_function_to_dict,
)

from ..dmrg import DMRG
from ..mpo import MatrixProductOperator
from ..mps import MatrixProductState
from ..quantum_algorithms.hamiltonian_simulation.trotterisation import TrotterSimulation


class ActiveSpaceSelection:
    def __init__(self, hamiltonian: dict[str, complex], coeff_matrix: ndarray):
        """Constructor

        Args:
            hamiltonian: System Hamiltonian
            coeff_matrix: HF coefficient matrix of shape (N, N)
        """
        self.hamiltonian = hamiltonian
        self.num_spin_orbitals = coeff_matrix.shape[0]
        self.num_orbitals = int(self.num_spin_orbitals / 2)
        self.coeff_matrix = coeff_matrix

    def run(
        self, num_active_orbitals: int, cost_function: Callable, **kwargs
    ) -> ndarray:
        """
        Perform active space selection by optimising a unitary transformation of the orbital coefficients.

        Args:
            num_active_orbitals [int]: Number of active orbitals to select
            cost_function [Callable]: The cost function to use for orbital optimisation
            kwargs: Valid arguments to provide -
                dmrg_max_mps_bond [int]: maximum bond dimension for DMRG, default 8
                dmrg_method [str]: either "one-site" or "two-site", default "two-site"
                dmrg_convergence_threshold [float]: convergence threshold for DMRG, default 1e-9
                dmrg_initial_state [MatrixProductState]: an initial MPS state for DMRG, default random MPS
                dmrg_maxiter: maximum number of sweeps to perform in DMRG, default 10
                cost_function_decay_power [float]: Required parameter for cost_mutual_info_decay, default 2.0

        Returns:
            Transformed coefficient matrix with optimal active orbitals
        """
        function_args = kwargs
        N = self.num_spin_orbitals
        assert self.coeff_matrix.shape[1] == N, "Coefficient matrix must be square"

        # Write the Hamiltonian and perfrom DMRG to get the initial state |psi>_C
        max_mps_bond = function_args.get("dmrg_max_mps_bond", 8)
        method = function_args.get("dmrg_method", "two-site")
        convergence_threshold = function_args.get("dmrg_convergence_threshold", 1e-9)
        initial_state = function_args.get("dmrg_initial_state", None)
        maxiter = function_args.get("dmrg_maxiter", 10)
        psi_C = self.run_dmrg(
            hamiltonian=self.hamiltonian,
            max_mps_bond=max_mps_bond,
            method=method,
            convergence_threshold=convergence_threshold,
            initial_state=initial_state,
            maxiter=maxiter,
        )

        # Cost function to MPO
        decay_power = function_args.get("cost_function_decay_power", 2.0)
        cost_mpo = self.build_cost_function_mpo(
            cost_function=cost_function, decay_power=decay_power
        )

        # Run BFGS optimisation to find optimal K
        theta_init = np.zeros((N**2,), dtype=float)  # Initial guess for theta
        theta_opt = self.optimise_K(theta_init, cost_mpo, psi_C)

        # Exponentiate K to get a unitary U = exp(K)
        K_opt = self.vector_to_antihermitian(theta_opt)
        U = self.exponentiate_K(K_opt)

        # Apply U to the input coefficient matrix, returning the transformed coefficient matrix (the new basis)
        self.transformed_coeff_matrix = self.coeff_matrix @ U

        return self.transformed_coeff_matrix

    def run_dmrg(
        self,
        hamiltonian: dict[str, complex],
        max_mps_bond: int | None,
        method: str,
        convergence_threshold: float,
        initial_state: MatrixProductState,
        maxiter: int,
    ) -> MatrixProductState:
        """Run DMRG to get an approximate groundstate in the initial MO basis"""
        dmrg = DMRG(
            hamiltonian,
            max_mps_bond=max_mps_bond,
            method=method,
            convergence_threshold=convergence_threshold,
            initial_state=initial_state,
        )
        _, psi = dmrg.run(maxiter=maxiter)
        return psi

    def build_cost_function_mpo(
        self, cost_function: Callable, decay_power: float
    ) -> MatrixProductOperator:
        """Build the cost function as an MPO"""
        d = cost_function_to_dict(
            cost_function, num_orbitals=self.num_orbitals, decay_power=decay_power
        )
        mpo = cost_function_dict_to_purity_mpo(self.num_spin_orbitals, d)
        return mpo

    def vector_to_antihermitian(self, theta: ndarray) -> ndarray:
        """
        Converts a real vector of length N^2 into an anti-Hermitian matrix K ∈ C^{N x N}.

        Diagonal entries are pure imaginary: iθ
        Off-diagonal: K[p,q] = a + ib, K[q,p] = -a + ib
        """
        norbs = self.num_spin_orbitals
        assert len(theta) == self.num_spin_orbitals**2, "theta must have length N^2"

        K = np.zeros((norbs, norbs), dtype=complex)
        idx = 0

        # Fill diagonals: all imaginary
        for i in range(norbs):
            K[i, i] = 1j * theta[idx]
            idx += 1

        # Fill upper triangle, set lower triangle with Hermitian conjugate
        for i in range(norbs):
            for j in range(i + 1, norbs):
                real = theta[idx]
                imag = theta[idx + 1]
                K[i, j] = real + 1j * imag
                K[j, i] = -real + 1j * imag  # = -conj(K[i,j])
                idx += 2

        return K

    def exponential_hopping_term(
        self, p: int, q: int, K_pq: complex
    ) -> MatrixProductOperator:
        """
        Construct the MPO for exp(K_pq * a_p† a_q - K_pq* * a_q† a_p).

        Args:
            p, q: Indices of orbitals (must be different)
            K_pq: Complex parameter

        Returns:
            MatrixProductOperator for exp(H), where H = K_pq * a_p† a_q - conj(K_pq) * a_q† a_p
        """
        if p == q:
            h_fermion = FermionOperator(((p, 1), (q, 0)), K_pq)

        else:
            # Build the FermionOperator, 1 is the creation operator, 0 is the annihilation operator
            # H = K_pq * a_p† a_q - conj(K_pq) * a_q† a_p
            h_fermion = FermionOperator(((p, 1), (q, 0)), K_pq) - FermionOperator(
                ((q, 1), (p, 0)), np.conj(K_pq)
            )

        # Map to QubitOperator using Jordan-Wigner
        h_qubit = jordan_wigner(h_fermion)

        # Convert to PauliwordOp
        h_pauli = PauliwordOp.from_openfermion(h_qubit, n_qubits=self.num_spin_orbitals)

        # Convert to a dictionary
        h_dict = h_pauli.to_dictionary
        h_dict = {k: v.real for k, v in h_dict.items()}

        # Create a circuit
        sim = TrotterSimulation(h_dict, duration=1.0, num_steps=1)
        qc = sim.circuit

        # Convert Qiskit circuit to MPO
        u_mpo = MatrixProductOperator.from_qiskit_circuit(qc)

        return u_mpo

    def build_trotterised_unitary(
        self, K: ndarray, trotter_steps: int = 1
    ) -> MatrixProductOperator:
        """
        Build an MPO approximation of the fermionic unitary:
            U = exp(Σ_{pq} K_{pq} a†_p a_q)

        using first-order Trotter decomposition.

        Args:
            K: Anti-Hermitian matrix (N x N)
            trotter_steps: Number of Trotter steps

        Returns:
            MatrixProductOperator representing the unitary
        """
        N = K.shape[0]
        assert K.shape[1] == N, "K must be square"
        assert np.allclose(K + K.conj().T, 0, atol=1e-10), "K must be anti-Hermitian"

        u_mpo = MatrixProductOperator.identity_mpo(N)
        dt = 1.0 / trotter_steps

        for _ in range(trotter_steps):
            for p in range(N):
                for q in range(p, N):
                    if abs(K[p, q]) > 1e-12:
                        K_dt = dt * K[p, q]
                        hop_exp_mpo = self.exponential_hopping_term(p, q, K_dt)
                        u_mpo = u_mpo * hop_exp_mpo

        return u_mpo

    def optimisation_cost(
        self, theta: ndarray, mpo: MatrixProductOperator, mps: MatrixProductState
    ) -> float:
        """Optimisation cost function.

        Args;
            theta: Paramter list for K
            mpo: QI cost function MPO
            mps: Groundstate approximation from DMRG

        Returns:
            < MPS | (exp(Σ_{pq} K_{pq} a†_p a_q))† MPO exp(Σ_{pq} K_{pq} a†_p a_q) | MPS >
        """
        # N = V.num_sites
        K = self.vector_to_antihermitian(theta)
        W_rotated = self.build_trotterised_unitary(K)  # returns an MPO
        transformed_state = mps.apply_mpo(W_rotated)
        doubled_transformed_state = transformed_state.to_two_copy_mps()
        cost = doubled_transformed_state.compute_expectation_value(mpo)
        return cost.real

    def optimise_K(
        self, theta_init: ndarray, mpo: MatrixProductOperator, mps: MatrixProductState
    ):
        """
        Run BFGS optimisation over K to minimise optimisation_cost.

        Args:
            theta_init: Initial guess for theta
            mpo: QI cost function MPO
            mps: Groundstate approximation from DMRG

        Returns:
            Optimal real-valued parameter vector θ defining anti-Hermitian K
        """
        # N = V.num_sites
        # num_params = N**2
        # theta0 = np.zeros(num_params)

        result = minimize(
            self.optimisation_cost,
            theta_init,
            args=(mpo, mps),
            method="COBYLA",
            options={"disp": True},
        )
        print("Optimisation result:", result.x)
        return result.x

    def exponentiate_K(self, K: ndarray) -> ndarray:
        """
        Compute U = exp(K) using eigendecomposition, where K is anti-Hermitian.

        Args:
        K: Anti-Hermitian matrix of shape (N, N)

        Returns:
        U = exp(K): a unitary matrix
        """
        assert K.shape[0] == K.shape[1], "K must be square"
        assert np.allclose(K + K.conj().T, 0), "K must be anti-Hermitian"

        # Eigendecomposition: K = V D V^{-1}
        eigvals, eigvecs = np.linalg.eig(K)

        # Compute exp(K) = V exp(D) V^{-1)
        exp_D = np.diag(np.exp(eigvals))
        V_inv = np.linalg.inv(eigvecs)
        U = eigvecs @ exp_D @ V_inv
        return U



from typing import List, TypeAlias, Union

# Underlying tensor objects can either be NumPy arrays or Sparse arrays
import numpy as np
import sparse
from numpy import ndarray
from numpy.linalg import svd
from qiskit.circuit import CircuitInstruction, Operation

# Qiskit quantum circuit integration
from qiskit.quantum_info import Operator
from sparse import SparseArray

DataOptions: TypeAlias = Union[ndarray, SparseArray]
QiskitOptions: TypeAlias = Union[CircuitInstruction, Operation]  # type: ignore


class Tensor:
    def __init__(
        self, data: DataOptions, indices: List[str], labels: List[str]
    ) -> None:
        """
        Constructor for the Tensor class.

        Args:
            data: The underlying data for the tensor. Internally this will always be stored as a sparse array.
            indices: Names for each dimension of the data array.
            labels: A list of labels to be associated with this tensor.
        """
        if isinstance(data, np.ndarray):
            sparse_array = sparse.COO(data)
        else:
            sparse_array = data
        self.data = sparse_array
        self.dimensions = data.shape
        self.rank = len(self.dimensions)
        self.indices = indices
        self.labels = labels

    @classmethod
    def from_array(
        cls, array: DataOptions, indices: List[str] = None, labels: List[str] = ["T1"]
    ) -> "Tensor":
        """
        Construct a tensor object from the array.

        Args:
            array: the underlying data for the tensor.
            index_prefix (optional): Default is "I".
            labels (optional): Default is "T1".

        Returns:
            tensor: The Tensor object.
        """

        data = array
        if array.size == 0:  # Check if the array is empty
            indices = []
        else:
            rank = len(array.shape)  # Use array.shape to calculate rank
            if not indices:
                index_prefix = "B"
                indices = [index_prefix + str(d_idx + 1) for d_idx in range(rank)]
            else:
                indices = indices
        labels = labels

        tensor = cls(data, indices, labels)

        return tensor

    @classmethod
    def from_qiskit_gate(
        cls,
        gate: QiskitOptions,
        indices: List[str] = None,
        labels: List[str] = ["T1"],
        dagger: bool = False,
    ) -> "Tensor":
        """
        Construct a tensor object from the array.

        Args:
            gate: the underlying qiskit object.
            indices (optional): Default is "I<int>" for inputs and "O<int>" for outputs.
            labels (optional): Default is "T1".
            dagger: If True will construct a tensor for the conjugate transpose of the gate

        Returns:
            tensor: The Tensor object.
        """

        if isinstance(gate, CircuitInstruction):
            gate = gate.operation

        num_qubits = gate.num_qubits
        num_dims = 2 * num_qubits
        shape = [2] * num_dims
        data = Operator(gate).reverse_qargs().data
        if dagger:
            data = data.conj().T
        data = np.reshape(data, shape)
        if not indices:
            indices = [0] * num_dims
            for idx in range(1, num_qubits + 1):
                indices[idx - 1] = "O" + str(idx)
                indices[num_qubits + idx - 1] = "I" + str(idx)
        labels = labels + [gate.name]

        tensor = cls(data, indices, labels)

        return tensor

    @classmethod
    def rank_3_copy(
        cls, indices: List[str] = ["B1", "R1", "L1"], labels: List[str] = ["T1"]
    ) -> "Tensor":
        """
        Construct a tensor object for the rank-3 copy tensor.

        Args:
            indices (optional): The list of indices. Defaults to ["B1", "R1", "L1"].
            labels (optional): The list of labels.

        Returns:
            A tensor.
        """
        array = np.array(
            [[[1, 0], [0, 1]], [[0, 0], [0, 1j * np.sqrt(2)]]], dtype=complex
        ).reshape(2, 2, 2)
        indices = indices
        labels = labels + ["copy3"]
        tensor = cls(array, indices, labels)
        return tensor

    @classmethod
    def rank_4_copy(
        cls, indices: List[str] = ["B1", "B2", "R1", "L1"], labels: List[str] = ["T1"]
    ) -> "Tensor":
        """
        Construct a tensor object for the rank-4 copy tensor.

        Args:
            indices (optional): The list of indices. Defaults to ["B1", "B2", "R1", "L1"].
            labels (optional): The list of labels.

        Returns:
            A tensor.
        """
        array = np.array(
            [
                [[[1, 0], [0, 1]], [[0, 0], [0, 0]]],
                [[[0, 0], [0, 0]], [[0, 0], [0, 1]]],
            ],
            dtype=complex,
        ).reshape(2, 2, 2, 2)
        indices = indices
        labels = labels + ["copy4"]
        tensor = cls(array, indices, labels)
        return tensor

    @classmethod
    def rank_3_copy_open(
        cls, indices: List[str] = ["B1", "R1", "L1"], labels: List[str] = ["T1"]
    ) -> "Tensor":
        """
        Construct a tensor object for the rank-3 copy tensor with open control.

        Args:
            indices (optional): The list of indices. Defaults to ["B1", "R1", "L1"].
            labels (optional): The list of labels.

        Returns:
            A tensor.
        """
        array = np.array(
            [[[1, 0], [0, 1]], [[1j * np.sqrt(2), 0], [0, 0]]], dtype=complex
        ).reshape(2, 2, 2)
        indices = indices
        labels = labels + ["copy3open"]
        tensor = cls(array, indices, labels)
        return tensor

    @classmethod
    def rank_4_copy_open(
        cls, indices: List[str] = ["B1", "B2", "R1", "L1"], labels: List[str] = ["T1"]
    ) -> "Tensor":
        """
        Construct a tensor object for the rank-4 copy tensor with open control.

        Args:
            indices (optional): The list of indices. Defaults to ["B1", "B2", "R1", "L1"].
            labels (optional): The list of labels.

        Returns:
            A tensor.
        """
        array = np.array(
            [
                [[[1, 0], [0, 1]], [[0, 0], [0, 0]]],
                [[[0, 0], [0, 0]], [[1, 0], [0, 0]]],
            ],
            dtype=complex,
        ).reshape(2, 2, 2, 2)
        indices = indices
        labels = labels + ["copy4open"]
        tensor = cls(array, indices, labels)
        return tensor

    @classmethod
    def rank_3_qiskit_gate(
        cls,
        gate: QiskitOptions,
        indices: List[str] = ["B1", "R1", "L1"],
        labels: List[str] = ["T1"],
    ) -> "Tensor":
        """
        Construct a tensor object for the rank-3 gate tensor.

        Args:
            indices (optional): The list of indices. Defaults to ["B1", "R1", "L1"].
            labels (optional): The list of labels.

        Returns:
            A tensor.
        """
        if isinstance(gate, CircuitInstruction):
            gate = gate.operation
        data = Operator(gate).reverse_qargs().data.reshape(2, 2)
        id_array = np.array([[1, 0], [0, 1]], dtype=complex).reshape(2, 2)
        array = np.array([id_array, (1j / np.sqrt(2)) * (id_array - data)]).reshape(
            2, 2, 2
        )
        indices = indices
        labels = labels + [f"rank3{gate.name}"]
        tensor = cls(array, indices, labels)
        return tensor

    @classmethod
    def rank_4_qiskit_gate(
        cls,
        gate: QiskitOptions,
        indices: List[str] = ["B1", "B2", "R1", "L1"],
        labels: List[str] = ["T1"],
    ) -> "Tensor":
        """
        Construct a tensor object for the rank-4 gate tensor.

        Args:
            indices (optional): The list of indices. Defaults to ["B1", "B2", "R1", "L1"].
            labels (optional): The list of labels.

        Returns:
            A tensor.
        """
        if isinstance(gate, CircuitInstruction):
            gate = gate.operation
        data = Operator(gate).reverse_qargs().data.reshape(2, 2)
        id_array = np.array([[1, 0], [0, 1]], dtype=complex).reshape(2, 2)
        zero_array = np.array([[0, 0], [0, 0]], dtype=complex).reshape(2, 2)
        array = np.array(
            [[id_array, zero_array], [zero_array, -0.5 * (data - id_array)]]
        ).reshape(2, 2, 2, 2)
        indices = indices
        labels = labels + [f"rank4{gate.name}"]
        tensor = cls(array, indices, labels)
        return tensor

    def __str__(self) -> str:
        """
        Defines output of print.
        """
        shape = str(self.dimensions)
        indices = str(self.indices)
        return f"Tensor with shape {shape} and indices {indices}"

    def reorder_indices(self, index_order: List[str]) -> None:
        """
        Used to change the order of indices in the tensor object.

        Args:
            index_order: The desired new ordering of indices.
        """
        old_indices = list(range(len(self.indices)))
        new_indices = [index_order.index(idx) for idx in self.indices]
        new_data = sparse.moveaxis(self.data, old_indices, new_indices)
        self.data = new_data
        self.indices = index_order
        self.dimensions = new_data.shape
        return

    def new_index_name(
        self, index_prefix: str = "B", num_new_indices: int = 1
    ) -> Union[str, List[str]]:
        """
        Generate a new index name not already in use.

        Args:
            index_prefix (optional): Default is "B".
            num_new_indices (optional): Number of new names required. Default is 1.

        Returns:
            The new index name. Returned as a str if num_new_indices=1, otherwise returned as List[str].
        """
        current_indices = [x for x in self.indices if len(x) > len(index_prefix)]
        current_vals = []
        for idx in current_indices:
            if (
                idx[: len(index_prefix)] == index_prefix
                and idx[len(index_prefix) :].isdigit()
            ):
                current_vals.append(int(idx[len(index_prefix) :]))
        if len(current_vals) > 0:
            max_current_val = max(current_vals)
        else:
            max_current_val = 0
        new_indices = [
            index_prefix + str(max_current_val + i)
            for i in range(1, num_new_indices + 1)
        ]

        if num_new_indices == 1:
            return new_indices[0]
        return new_indices

    def get_dimension_of_index(self, index_name: str) -> int:
        """
        Get the dimension of an index of the tensor.

        Args:
            index_name: The name of the index.

        Returns:
            The dimension associated to index_name.
        """
        return self.dimensions[self.indices.index(index_name)]

    def get_total_dimension_of_indices(self, idxs: List[str]) -> int:
        """
        Get the total dimension of a list of indices of the tensor.

        Args:
            idxs: The names of the indices.

        Returns:
            The product of dimensions associated to each index in idxs.
        """
        dims = [self.get_dimension_of_index(idx) for idx in idxs]
        total = np.prod(dims)
        return total

    def combine_indices(self, idxs: List[str], new_index_name: str = None) -> None:
        """
        Merge two or more indices together in the tensor.

        Args:
            idxs: The indices to merge.
            new_index_name (optional): What to call the resulting merged index. Defaults to a new index name.
        """
        original_index_ordering = self.indices
        combined_index_dim = self.get_total_dimension_of_indices(idxs)

        temp_index_ordering = [idx for idx in idxs]
        temp_shape = [self.get_dimension_of_index(idx) for idx in idxs]
        for idx in original_index_ordering:
            if idx not in idxs:
                temp_index_ordering.append(idx)
                temp_shape.append(self.get_dimension_of_index(idx))
        self.reorder_indices(temp_index_ordering)

        new_shape = [combined_index_dim] + temp_shape[len(idxs) :]
        new_data = sparse.reshape(self.data, new_shape)
        if not new_index_name:
            new_index_name = self.new_index_name()
        new_index_ordering = [new_index_name] + temp_index_ordering[len(idxs) :]
        new_rank = len(new_index_ordering)

        self.data = new_data
        self.indices = new_index_ordering
        self.dimensions = tuple(new_shape)
        self.rank = new_rank

        return

    def tensor_to_matrix(self, input_idxs: List[str], output_idxs: List[str]) -> None:
        """
        Reshape the tensor into a matrix.

        Args:
            input_idxs: The indices to be treated as matrix inputs.
            output_idxs: The indices to be treated as matrix outputs.
        """
        if len(input_idxs) > 0:
            self.combine_indices(input_idxs, new_index_name="I1")
        if len(output_idxs) > 0:
            self.combine_indices(output_idxs, new_index_name="O1")
        return

    def multiply_by_constant(self, const: complex) -> None:
        """
        Multiply the tensor by a constant.

        Args:
            const: The constant to multiply by.
        """
        self.data = self.data * const
        return

    def dagger(self) -> None:
        """
        Get the conjugate transpose
        """
        self.data = self.data.conj()
        return

    def get_closest_unitary(
        self, input_indices: list[str], output_indices: list[str]
    ) -> "Tensor":
        """
        Perform a polar decomposition to determine the closest unitary matrix to the given tensor.

        Args:
            input_indices: The indices to treat as the input indices for the matrix representation
            output_indices: The indices to treat as the output indices for the matrix representation

        Returns:
            A new tensor that is unitary as a matrix with index ordering same as self
        """
        matrix = self.tensor_to_matrix(input_indices, output_indices)
        u, _, vh = svd(matrix, full_matrices=False)
        unitary_part = u @ vh
        input_dims = [self.get_dimension_of_index(i) for i in input_indices]
        output_dims = [self.get_dimension_of_index(i) for i in output_indices]
        shape = tuple(input_dims + output_dims)
        unitary_part = unitary_part.reshape(
            shape
        )  # This will have index ordering output_inds, input_inds
        new_t = Tensor(unitary_part, output_indices + input_indices, labels=self.labels)
        new_t.reorder_indices(self.indices)
        return new_t



import copy
import re

import numpy as np
import scipy
import scipy.optimize
from numpy import ndarray
from numpy.linalg import eig, svd
from qiskit import QuantumCircuit
from qiskit.circuit.library import UnitaryGate

from ..fidelity_metrics import hilbert_schmidt_inner_product
from ..mpo import MatrixProductOperator
from ..quantum_algorithms.variational.ansatz_circuits import (
    identity_brickwork_circuit,
)
from ..tensor import Tensor
from ..tn import TensorNetwork
from .utils import kak_recomposition


class ApproximateDiagonalisation:
    """
    A class for approximately diagonalising an MPO
    """

    def __init__(self, mpo: MatrixProductOperator, num_layers: int) -> None:
        """
        Class constructor.

        Args:
            mpo: The MPO that will be appoximately diagonalised
            num_layers: The number of layers to use in the ansatz circuit
        """
        self.reference = MatrixProductOperator.from_diagonal_matrix(
            [-10] * 2 + [10] * 14
        )
        self.qc = identity_brickwork_circuit(mpo.num_sites, num_layers)
        self.num_qubits = mpo.num_sites
        self.mpo_to_diag = mpo
        self.set_ansatz()
        self.set_ansatz_dag()
        self.set_default_indices()
        self.build_tn()
        self.approximately_diagonalised_mpo = (
            self.construct_approximately_diagonalised_mpo()
        )

    def set_default_indices(self) -> None:
        """
        Set default indices for the full TN
        """

        def _index_splitter(idx):
            """Split the TN index into QW<number>, N<number>"""
            match = re.match(r"(QW\d+)(N\d+)", idx)
            qw, n = match.groups()
            return qw, n

        def _get_left_tn_indices():
            """Get all TN indices with N number 0"""
            left_tn_indices = [0] * self.num_qubits
            for t in self.ansatz.tensors:
                for idx in t.indices:
                    qw, n = _index_splitter(idx)
                    if n[1:] == "0":
                        left_tn_indices[int(qw[2:]) - 1] = idx
            return left_tn_indices

        def _get_right_tn_indices():
            """Get all TN indices with maximum N number for each QW number"""
            index_dict = {f"QW{x}": [] for x in range(1, self.num_qubits + 1)}
            for t in self.ansatz.tensors:
                for idx in t.indices:
                    qw, n = _index_splitter(idx)
                    index_dict[qw].append(int(n[1:]))
            right_tn_tensors = []
            for k, v in index_dict.items():
                max_n_number = max(v)
                index = k + "N" + str(max_n_number)
                right_tn_tensors.append(index)
            return right_tn_tensors

        left_tn_indices = _get_left_tn_indices()
        right_tn_indies = _get_right_tn_indices()

        for t in self.ansatz_dag.tensors:
            original_t_indices = t.indices
            new_t_indices = []
            for idx in original_t_indices:
                qw, _ = _index_splitter(idx)
                if idx in left_tn_indices:
                    new_t_indices.append(f"T{qw[2:]}")
                elif idx in right_tn_indies:
                    new_t_indices.append(f"V{qw[2:]}")
                else:
                    new_t_indices.append(idx)
            t.indices = new_t_indices
        self.mpo_to_diag.set_default_indices(
            internal_prefix="A", input_prefix="V", output_prefix="W"
        )
        for t in self.ansatz.tensors:
            original_t_indices = t.indices
            new_t_indices = []
            for idx in original_t_indices:
                qw, _ = _index_splitter(idx)
                if idx in left_tn_indices:
                    new_t_indices.append(f"W{qw[2:]}")
                elif idx in right_tn_indies:
                    new_t_indices.append(f"X{qw[2:]}")
                else:
                    new_t_indices.append(idx + "_")
            t.indices = new_t_indices
        self.reference.set_default_indices(
            internal_prefix="B", input_prefix="X", output_prefix="T"
        )
        return

    def update_circuit(self, variational_index: int, optimal_update: ndarray) -> None:
        """
        Update the quantum circuit with the optimal local update

        Args:
            variational_index: The local index to be updated
            optimal_value: The optimal update array
        """
        new_inst = UnitaryGate(optimal_update)
        qidxs = [
            self.qc.data[variational_index - 1].qubits[x]._index
            for x in range(len(self.qc.data[variational_index - 1].qubits))
        ]
        self.qc.data[variational_index - 1] = (new_inst, qidxs[::-1], [])
        return

    def set_ansatz(self) -> None:
        """Update ansatz after circuit update"""
        self.ansatz = TensorNetwork.from_qiskit_circuit(self.qc)
        site_idx = 1
        for t in self.ansatz.tensors:
            if len(t.indices) == 4:
                t.labels.append(f"variational_site_{site_idx}")
                site_idx += 1
        self.num_variational_parameters = site_idx - 1
        return

    def set_ansatz_dag(self) -> None:
        """Update ansatz dag after circuit update"""
        self.ansatz_dag = TensorNetwork.from_qiskit_circuit(self.qc, dagger=True)
        site_idx = 1
        for t in self.ansatz_dag.tensors[::-1]:
            if len(t.indices) == 4:
                t.labels.append(f"variational_site_{site_idx}")
                site_idx += 1
        return

    def build_tn(self) -> None:
        """Build the full tensor network assuming indices have been set"""
        self.tn = TensorNetwork(
            self.reference.tensors
            + self.ansatz_dag.tensors
            + self.mpo_to_diag.tensors
            + self.ansatz.tensors,
            name="ApproximateDiagonalisation",
        )
        return

    def construct_approximately_diagonalised_mpo(self) -> "MatrixProductOperator":
        """
        Construct the approximately diagonalised MPO
        """
        mpo = copy.deepcopy(self.mpo_to_diag)
        ansatz_mpo = MatrixProductOperator.from_qiskit_circuit(self.qc)
        ansatz_dag_mpo = MatrixProductOperator.from_qiskit_circuit(self.qc.inverse())

        mpo = ansatz_mpo * mpo
        mpo = mpo * ansatz_dag_mpo

        return mpo

    def get_local_indices(self, variational_idx: int) -> tuple[list[str], list[str]]:
        """
        For the local site get the expected indices of the environment tensor

        Args:
            variational_idx: The index of the current ansatz site

        Returns:
            output_inds, input_inds for the environment tensor
        """
        ansatz_tensor = self.ansatz.get_tensors_from_label(
            f"variational_site_{variational_idx}"
        )[0]
        ansatz_dag_tensor = self.ansatz_dag.get_tensors_from_label(
            f"variational_site_{variational_idx}"
        )[0]
        input_inds = ansatz_tensor.indices
        output_inds = ansatz_dag_tensor.indices
        return output_inds, input_inds

    def form_environment_matrix(self, variational_idx: int) -> ndarray:
        """
        Form the environment matrix for a local variational tensor

        Args:
            variational_idx: The index of the current ansatz site

        Returns:
            A matrix for the environment
        """
        tn_copy = copy.deepcopy(self.tn)
        site_label = f"variational_site_{variational_idx}"
        tn_copy.pop_tensors_by_label([site_label])
        env_tensor = tn_copy.contract_entire_network()
        env_copy = copy.deepcopy(env_tensor)
        output_inds, input_inds = self.get_local_indices(variational_idx)
        env_copy.tensor_to_matrix(input_inds, output_inds)
        env_mat = env_copy.data.todense()
        return env_mat

    def get_maximum_eigenvector(self, mat: ndarray) -> ndarray:
        """
        For a given matrix, get the eigenvector associated to the maximum eigenvalue

        Args:
            mat: The input matrix

        Returns:
            The maximum eigenvector
        """
        evals, evecs = eig(mat)
        max_eval = max(evals)
        max_eval_idx = list(evals).index(max_eval)
        max_evec = evecs[:, max_eval_idx]
        return max_evec

    def get_closest_unitary(self, mat: ndarray) -> ndarray:
        """
        Get the closest unitary to a given matrix

        Args:
            mat: The input matrix

        Returns:
            The closest unitary to mat under Frobenius norm
        """
        u, _, vh = svd(mat, full_matrices=False)
        unitary_part = u @ vh
        return unitary_part

    def quadratic_optimisation_over_unitaries(self, mat: ndarray) -> ndarray:
        """
        Perform a quadratic optimisation over the unitary group with the given matrix

        Args:
            mat: A matrix

        Returns:
            The optimised unitary
        """

        def _cost_function(params):
            uni = kak_recomposition(
                params[:3], params[3:6], params[6:9], params[9:12], params[12:]
            )
            # uni = symmetry_preserving_two_qubit_gate(params[0], params[1:4], params[-1])
            # uni = givens_rotation(params[0])
            uni_vec = uni.reshape((16,), order="F")
            quad = uni_vec.conj().T @ mat @ uni_vec
            return -1.0 * quad.real

        initial_params = [0.0] * 15
        # initial_params = [0.0] * 5
        # initial_params = [0.0]
        method = "COBYLA"
        # bounds = ([(0.0, 2*np.pi)])
        # bounds = ([(0.0, 2*np.pi) for _ in range(5)])
        bounds = (
            [(0.0, 2 * np.pi) for _ in range(6)]
            + [(0.0, np.pi / 4) for _ in range(3)]
            + [(0.0, 2 * np.pi) for _ in range(6)]
        )
        result = scipy.optimize.minimize(
            _cost_function, initial_params, method=method, bounds=bounds
        )
        optimised_params = result.x
        # optimised_uni = givens_rotation(optimised_params[0])
        # optimised_uni = symmetry_preserving_two_qubit_gate(optimised_params[0], optimised_params[1:4], optimised_params[-1])
        optimised_uni = kak_recomposition(
            optimised_params[:3],
            optimised_params[3:6],
            optimised_params[6:9],
            optimised_params[9:12],
            optimised_params[12:],
        )
        return optimised_uni

    def local_update(self, variational_index: int) -> None:
        """
        Perform a local optimisation at the given index

        Args:
            variational_index: The index of the current local site
        """
        env_mat = self.form_environment_matrix(variational_index)
        new_site_data = self.quadratic_optimisation_over_unitaries(env_mat)

        # local_tensor = self.ansatz.tensors[variational_index - 1]
        # dim = int(2 ** (len(local_tensor.indices) / 2))

        # max_evec = self.get_maximum_eigenvector(env_mat)
        # new_site_data = max_evec.reshape((dim, dim))
        # new_site_data = self.get_closest_unitary(new_site_data)

        self.update_circuit(variational_index, new_site_data)
        self.set_ansatz()
        self.set_ansatz_dag()
        self.set_default_indices()
        self.build_tn()

        return

    def run(self, num_sweeps: int = 10) -> QuantumCircuit:
        """
        Optimise the ansatz to approximately diagonalise the given MPO

        Args:
            num_sweeps: The number of sweeps to perform

        Returns:
            The optimised quantum circuit
        """
        for _ in range(num_sweeps):
            for idx in range(1, self.num_variational_parameters + 1):
                self.local_update(idx)
            for idx in list(range(1, self.num_variational_parameters))[::-1]:
                self.local_update(idx)
        self.approximately_diagonalised_mpo = (
            self.construct_approximately_diagonalised_mpo()
        )
        return self.qc


class ApproximateDiagonalisationMPO:
    """
    A class for approximately diagonalising an MPO
    """

    def __init__(self, mpo: MatrixProductOperator, max_bond: int) -> None:
        """
        Class constructor.

        Args:
            mpo: The MPO that will be appoximately diagonalised
            max_bond: The maximum allowed bond dimension
        """
        reference = MatrixProductOperator.from_increasing_diagonal_matrix(mpo.num_sites)
        self.ansatz = MatrixProductOperator.random_mpo(mpo.num_sites, max_bond)
        self.mpo_to_diag = mpo
        self.max_bond = max_bond
        self.ansatz_dag = copy.deepcopy(self.ansatz)
        self.ansatz_dag.dagger()
        for t in self.ansatz.tensors:
            t.labels.append("variational")
            t.labels.append(f"variational_site_{self.ansatz.tensors.index(t)+1}")
        for t in self.ansatz_dag.tensors:
            t.labels.append(f"variational_site_{self.ansatz_dag.tensors.index(t)+1}")
        reference.set_default_indices(
            internal_prefix="A", input_prefix="T", output_prefix="V"
        )
        self.ansatz_dag.set_default_indices(
            internal_prefix="B", input_prefix="V", output_prefix="W"
        )
        self.mpo_to_diag.set_default_indices(
            internal_prefix="C", input_prefix="W", output_prefix="X"
        )
        self.ansatz.set_default_indices(
            internal_prefix="D", input_prefix="X", output_prefix="T"
        )
        self.tn = TensorNetwork(
            reference.tensors
            + self.ansatz_dag.tensors
            + self.mpo_to_diag.tensors
            + self.ansatz.tensors,
            name="ApproximateDiagonalisation",
        )
        self.reference = reference
        self.approximately_diagonalised_mpo = (
            self.construct_approximately_diagonalised_mpo()
        )

    def construct_approximately_diagonalised_mpo(self) -> "MatrixProductOperator":
        """
        Construct the approximately diagonalised MPO
        """
        adag = copy.deepcopy(self.ansatz_dag)
        mpo = copy.deepcopy(self.mpo_to_diag)
        a = copy.deepcopy(self.ansatz)
        approximately_diagonalised_mpo = adag * mpo
        approximately_diagonalised_mpo = approximately_diagonalised_mpo * a
        return approximately_diagonalised_mpo

    def get_local_indices(self, variational_idx: int) -> tuple[list[str], list[str]]:
        """
        For the local site get the expected indices of the environment tensor

        Args:
            variational_idx: The index of the current ansatz site

        Returns:
            output_inds, input_inds for the environment tensor
        """
        ansatz_tensor = self.ansatz.tensors[variational_idx - 1]
        ansatz_dag_tensor = self.ansatz_dag.tensors[variational_idx - 1]
        input_inds = ansatz_tensor.indices
        output_inds = ansatz_dag_tensor.indices
        return output_inds, input_inds

    def form_environment_matrix(self, variational_idx: int) -> ndarray:
        """
        Form the environment matrix for a local variational tensor

        Args:
            variational_idx: The index of the current ansatz site

        Returns:
            A matrix for the environment
        """
        tn_copy = copy.deepcopy(self.tn)
        site_label = f"variational_site_{variational_idx}"
        tn_copy.pop_tensors_by_label([site_label])
        env_tensor = tn_copy.contract_entire_network()
        env_copy = copy.deepcopy(env_tensor)
        output_inds, input_inds = self.get_local_indices(variational_idx)
        env_copy.tensor_to_matrix(input_inds, output_inds)
        env_mat = env_copy.data.todense()
        return env_mat

    def get_maximum_eigenvector(self, mat: ndarray) -> ndarray:
        """
        For a given matrix, get the eigenvector associated to the maximum eigenvalue

        Args:
            mat: The input matrix

        Returns:
            The maximum eigenvector
        """
        evals, evecs = eig(mat)
        max_eval = max(evals)
        max_eval_index = list(evals).index(max_eval)
        max_evec = evecs[:, max_eval_index]
        return max_evec

    def get_closest_isometry(self, data: ndarray) -> ndarray:
        """
        The global unitary needs to be unitary. We can achieve this by ensuring each local update is an isometry.

        Args:
            data: The optimised data for a local update

        Returns:
            The data for the closest isometry
        """
        u, _, vh = svd(data, full_matrices=False)
        new_data = u @ vh
        return new_data

    def local_update(self, variational_index: int) -> None:
        """
        Perform a local optimisation at the given index

        Args:
            variational_index: The index of the current local site
        """
        local_tensor = self.ansatz.tensors[variational_index - 1]
        local_indices = local_tensor.indices
        local_dims = [local_tensor.get_dimension_of_index(idx) for idx in local_indices]
        local_labels = local_tensor.labels

        env_mat = self.form_environment_matrix(variational_index)
        max_evec = self.get_maximum_eigenvector(env_mat)
        new_site_data = max_evec.reshape(tuple(local_dims))
        new_site_data = self.get_closest_isometry(new_site_data)
        new_tensor = Tensor(new_site_data, indices=local_indices, labels=local_labels)
        site_label = f"variational_site_{variational_index}"
        self.ansatz.pop_tensors_by_label([site_label])
        self.ansatz.add_tensor(new_tensor, variational_index - 1)
        self.update_ansatz_dag()
        return

    def check_global_unitarity(self) -> None:
        """
        The output MPO will be a scaled version of a unitary. Here we enforce full unitarity.
        """
        # Check the magnitude
        ansatz = copy.deepcopy(self.ansatz)
        ansatz_copy = copy.deepcopy(self.ansatz)
        ip = hilbert_schmidt_inner_product(ansatz, ansatz_copy).real
        scale_factor = np.sqrt(ip / (2**self.ansatz.num_sites))
        self.ansatz.multiply_by_constant(1 / scale_factor)
        self.ansatz_dag.multiply_by_constant(1 / scale_factor)
        return

    def update_ansatz_dag(self) -> None:
        """
        Update ansatz_dag after changes to ansatz.
        """
        self.ansatz_dag = copy.deepcopy(self.ansatz)
        self.ansatz_dag.dagger()
        self.ansatz_dag.set_default_indices(
            internal_prefix="B", input_prefix="V", output_prefix="W"
        )
        return

    def update_tn(self) -> None:
        """
        Update tn after changes to ansatz.
        """
        self.tn = TensorNetwork(
            self.reference.tensors
            + self.ansatz_dag.tensors
            + self.mpo_to_diag.tensors
            + self.ansatz.tensors,
            name="ApproximateDiagonalisation",
        )
        return

    def run(self, num_sweeps: int = 10) -> MatrixProductOperator:
        """
        Optimise the ansatz to approximately diagonalise the given MPO

        Args:
            num_sweeps: The number of sweeps to perform

        Returns:
            The optimised ansatz
        """
        for _ in range(num_sweeps):
            for idx in range(1, self.ansatz.num_sites + 1):
                self.local_update(idx)
                self.update_tn()
            for idx in list(range(1, self.ansatz.num_sites + 1))[::-1]:
                self.local_update(idx)
                self.update_tn()
        self.check_global_unitarity()
        self.approximately_diagonalised_mpo = (
            self.construct_approximately_diagonalised_mpo()
        )
        return self.ansatz



import copy
import re

import numpy as np
from numpy import ndarray
from numpy.linalg import svd
from qiskit import QuantumCircuit
from qiskit.circuit.library import UnitaryGate
from scipy.linalg import null_space, polar

from ..circuit_simulator import CircuitSimulator
from ..fidelity_metrics import state_uhlmann_fidelity
from ..mps import MatrixProductState
from ..tensor import Tensor
from ..tn import TensorNetwork


class MPSOptimiser:
    """
    A class for locally optimising a quantum circuit with respect to a reference MPS and the HS distance
    """

    def __init__(self, qc: QuantumCircuit, reference: MatrixProductState) -> None:
        """
        Constructor

        Args:
            qc: The quantum circuit that will be optimised
            reference: The reference MPS
        """
        self.qc = qc
        self.reference = reference
        self.num_qubits = qc.num_qubits
        self.set_tn()
        self.error = self.calculate_error()
        self.fidelity = self.get_fidelity()
        self.optimisation_dict = {
            "optimisation_iteration": [0],
            "error": [self.error],
            "fidelity": [self.fidelity],
        }

    def get_tn_external_indices(self, tn: TensorNetwork) -> tuple[list[str], list[str]]:
        """
        Get the left and right indices of TN
        """

        def _index_splitter(idx):
            """Split the TN index into QW<number>, N<number>"""
            match = re.match(r"(QW\d+)(N\d+)", idx)
            qw, n = match.groups()
            return qw, n

        def _get_left_tn_indices():
            """Get all TN indices with N number 0"""
            left_tn_indices = [0] * self.num_qubits
            for t in tn.tensors:
                for idx in t.indices:
                    qw, n = _index_splitter(idx)
                    if n[1:] == "0":
                        left_tn_indices[int(qw[2:]) - 1] = idx
            return left_tn_indices

        def _get_right_tn_indices():
            """Get all TN indices with maximum N number for each QW number"""
            index_dict = {f"QW{x}": [] for x in range(1, self.num_qubits + 1)}
            for t in tn.tensors:
                for idx in t.indices:
                    qw, n = _index_splitter(idx)
                    index_dict[qw].append(int(n[1:]))
            right_tn_tensors = []
            for k, v in index_dict.items():
                max_n_number = max(v)
                index = k + "N" + str(max_n_number)
                right_tn_tensors.append(index)
            return right_tn_tensors

        left_tn_indices = _get_left_tn_indices()
        right_tn_indies = _get_right_tn_indices()
        return left_tn_indices, right_tn_indies

    def apply_initial_state(self):
        """
        Apply the all zero initial state to tn to form a state
        """
        for idx in self.left_tn_indices:
            zero_data = np.array([1, 0], dtype=complex).reshape((2,))
            zero_tensor = Tensor(zero_data, [idx], ["Zero"])
            self.tn.add_tensor(zero_tensor)
        return

    def ip_rr(self) -> complex:
        """
        Calculate <R|R> where R is the reference MPS
        """
        return 1.0 + 0.0j

    def ip_tr(self) -> complex:
        """
        Calculate <T|R> where R is the reference MPS and T is the TN
        """
        return self.ip_rt().conjugate()

    def ip_rt(self) -> complex:
        """
        Calculate <R|T> where R is the reference MPS and T is the TN
        """
        tn = self.build_ip_rt_tn()
        ip = tn.contract_entire_network()
        return ip

    def ip_tt(self) -> complex:
        """
        Calculate <T|T> where T is the TN
        """
        return 1.0 + 0.0j

    def calculate_error(self) -> float:
        """
        Calculate the squared Frobenius norm between the reference MPS and the TN
        """
        err = self.ip_rr() - self.ip_rt() - self.ip_tr() + self.ip_tt()
        return max(err.real, 0.0)  # It will be real anyway

    def get_fidelity(self) -> float:
        """
        Get the fidelity
        """
        overlap = self.ip_tr()
        fid = np.abs(overlap) ** 2
        return fid

    def build_ip_rt_tn(self) -> TensorNetwork:
        def _index_splitter(idx):
            """Split the TN index into QW<number>, N<number>"""
            match = re.match(r"(QW\d+)(N\d+)", idx)
            qw, n = match.groups()
            return qw, n

        r = copy.deepcopy(self.reference)
        r.dagger()
        r.set_default_indices("A", "T")
        tn = copy.deepcopy(self.tn)
        for t in tn.tensors:
            original_t_indices = t.indices
            new_t_indices = []
            for idx in original_t_indices:
                qw, _ = _index_splitter(idx)
                if idx in self.right_tn_indices:
                    new_t_indices.append(f"T{qw[2:]}")
                else:
                    new_t_indices.append(idx)
            t.indices = new_t_indices

        full_tn = TensorNetwork(r.tensors + tn.tensors)
        return full_tn

    def get_environment_vector(self, variational_index: int) -> ndarray:
        tn = self.build_ip_rt_tn()
        site_label = f"variational_site_{variational_index}"
        popped_t = tn.pop_tensors_by_label([site_label])
        env_tensor = tn.contract_entire_network()
        env_copy = copy.deepcopy(env_tensor)
        output_inds = popped_t[0].indices
        env_copy.combine_indices(output_inds)
        env_vec = env_copy.data.todense()
        return env_vec

    def get_closest_unitary(self, mat: ndarray) -> ndarray:
        """
        Get the closest unitary to a given matrix

        Args:
            mat: The input matrix

        Returns:
            The closest unitary to mat under Frobenius norm
        """
        u, _, vh = svd(mat, full_matrices=False)
        unitary_part = u @ vh
        return unitary_part

    def update_circuit(self, variational_index: int, optimal_update: ndarray) -> None:
        """
        Update the quantum circuit with the optimal local update

        Args:
            variational_index: The local index to be updated
            optimal_value: The optimal update array
        """
        new_inst = UnitaryGate(optimal_update)
        qidxs = [
            self.qc.data[variational_index - 1].qubits[x]._index
            for x in range(len(self.qc.data[variational_index - 1].qubits))
        ]
        self.qc.data[variational_index - 1] = (new_inst, qidxs[::-1], [])
        return

    def set_tn(self) -> None:
        """
        Reset TN after changes to qc
        """
        self.tn = TensorNetwork.from_qiskit_circuit(self.qc)
        for t in self.tn.tensors:
            t.labels.append(f"variational_site_{self.tn.tensors.index(t)+1}")
        self.num_variational_sites = len(self.tn.tensors)
        self.left_tn_indices, self.right_tn_indices = self.get_tn_external_indices(
            self.tn
        )
        self.apply_initial_state()
        return

    def local_update(self, variational_index: int) -> None:
        """
        Perform a local update

        Args:
            variational_index: The index of the current site
        """
        site_index = f"variational_site_{variational_index}"
        local_tensor = self.tn.get_tensors_from_label(site_index)[0]
        local_indices = local_tensor.indices
        local_dimensions = [
            local_tensor.get_dimension_of_index(idx) for idx in local_indices
        ]
        dim = np.prod(local_dimensions)

        env_vec = self.get_environment_vector(variational_index)
        update = env_vec
        update = update.reshape((int(np.sqrt(dim)), int(np.sqrt(dim))))
        unitary_update = self.get_closest_unitary(update)
        self.update_circuit(variational_index, unitary_update)
        self.set_tn()
        return

    def run(self, num_sweeps: int = 10) -> QuantumCircuit:
        """
        Optimise the ansatz to match the reference

        Args:
            num_sweeps: The number of sweeps to perform

        Returns:
            The optimised quantum circuit
        """
        for it_number in range(num_sweeps):
            for idx in range(1, len(self.qc.data) + 1):
                self.local_update(idx)
            for idx in list(range(1, len(self.qc.data) + 1))[::-1]:
                self.local_update(idx)
            self.error = self.calculate_error()
            self.fidelity = self.get_fidelity()
            self.optimisation_dict["optimisation_iteration"].append(it_number + 1)
            self.optimisation_dict["error"].append(self.error)
            self.optimisation_dict["fidelity"].append(self.fidelity)
        return self.qc


class MPSAnalyticDecomposition:
    """A class to analytically decompose MPS as quantum circuits"""

    def __init__(
        self, mps: MatrixProductState, max_layers: int, target_fidelity: float
    ):
        """Constructor

        Args:
            mps: The MPS to map to a quantum circuit
            max_layers: The maximum number of allowed staircase circuit layers
            target_fidelity: The target fidelity between the quantum circuit and the MPS
        """
        self.mps = mps
        self.num_sites = self.mps.num_sites
        self.max_layers = max_layers
        self.target_fidelity = target_fidelity
        self.qc = QuantumCircuit(mps.num_sites)
        self.num_layers = 0
        self.fidelity = 0.0

    def compress_to_bond_dim_2(self, mps: MatrixProductState) -> MatrixProductState:
        """Compress the current mps to bond dimension 2"""
        mps_copy = copy.deepcopy(mps)
        mps_copy.compress(2)
        mps_copy.normalise()
        return mps_copy

    def extend_to_unitary(
        self, tensor: Tensor, position: str | None = None
    ) -> np.ndarray:
        """Constructs a unitary matrix from a given tensor"""
        data = copy.deepcopy(tensor)

        # Determine reshape based on position
        if position == "first":
            data.reorder_indices([data.indices[1], data.indices[0]])
            matrix = data.data.todense().reshape((4, 1))
        elif position == "last":
            data.reorder_indices([data.indices[1], data.indices[0]])
            matrix = data.data.todense().reshape((2, 2))
        else:
            data.tensor_to_matrix(
                [tensor.indices[0]], [tensor.indices[2], tensor.indices[1]]
            )
            matrix = data.data.todense()

        shape = matrix.shape

        orthogonal_basis_1 = null_space(matrix)
        orthogonal_basis_2 = null_space(matrix.conj().T)

        if shape[0] > shape[1]:
            unitary = np.concatenate([matrix, orthogonal_basis_2], 1)
        elif shape[0] < shape[1]:
            unitary = np.concatenate([matrix.conj().T, orthogonal_basis_1], 1).conj().T
        else:
            unitary = matrix

        unitary, _ = polar(unitary)
        return unitary

    def bond_dim_2_to_qc_exact(
        self, bond_dim_2_mps: MatrixProductState
    ) -> QuantumCircuit:
        """Map a bond dimension 2 MPS to a quantum circuit exactly"""

        mps = bond_dim_2_mps
        mps.move_orthogonality_centre(1)

        mps_dims = [mps.tensors[idx].dimensions[0] for idx in range(1, mps.num_sites)]
        bond_dim_1_idxs = (
            [0] + [i + 1 for i, x in enumerate(mps_dims) if x == 1] + [mps.num_sites]
        )
        separate_mps_arrays = []
        for i in range(len(bond_dim_1_idxs) - 1):
            separate_mps_arrays.append(
                [
                    mps.tensors[idx].data.todense()
                    for idx in list(range(mps.num_sites))[
                        bond_dim_1_idxs[i] : bond_dim_1_idxs[i + 1]
                    ]
                ]
            )

        separate_mps = []

        for arrays in separate_mps_arrays:
            if len(arrays) == 1:
                array = copy.deepcopy(arrays[0])
                array = array.reshape((2,))
                separate_mps.append(MatrixProductState.from_arrays([array]))
                continue
            elif len(arrays) == 2:
                first_array = copy.deepcopy(arrays[0])
                first_array = first_array.reshape((2, 2))
                last_array = copy.deepcopy(arrays[1])
                last_array = last_array.reshape((2, 2))
                separate_mps.append(
                    MatrixProductState.from_arrays([first_array, last_array])
                )
                continue

            reshaped_arrays = []
            first_array = copy.deepcopy(arrays[0])
            if first_array.ndim == 2:
                pass
            else:
                first_array = first_array.reshape(
                    (first_array.shape[1], first_array.shape[2])
                )
            reshaped_arrays.append(first_array)
            for i in range(1, len(arrays) - 1):
                array = copy.deepcopy(arrays[i])
                reshaped_arrays.append(array)
            last_array = copy.deepcopy(arrays[-1])
            if last_array.ndim == 2:
                pass
            else:
                last_array = last_array.reshape(
                    (last_array.shape[0], last_array.shape[2])
                )
            reshaped_arrays.append(last_array)
            separate_mps.append(MatrixProductState.from_arrays(reshaped_arrays))

        qcs = []
        qidxs = []
        for sub_mps in separate_mps:
            if sub_mps.num_sites == 1:
                v = sub_mps.tensors[0].data.todense().reshape((2,))
                a, b = v
                v_perp = np.array([-np.conj(b), np.conj(a)])
                unitary = np.column_stack((v, v_perp))
                gate = UnitaryGate(unitary)
                qc = QuantumCircuit(1)
                qc.append(gate, [0])
                qcs.append(qc)
                if len(qidxs) == 0:
                    qidxs.append([0])
                else:
                    qidxs.append([qidxs[-1][-1] + 1])
                continue

            unitaries = []
            first_uni = self.extend_to_unitary(sub_mps.tensors[0], "first")
            unitaries.append(first_uni)
            for tidx in range(1, sub_mps.num_sites - 1):
                t = sub_mps.tensors[tidx]
                uni = self.extend_to_unitary(t)
                unitaries.append(uni)
            final_uni = self.extend_to_unitary(sub_mps.tensors[-1], "last")
            unitaries.append(final_uni)

            qc = QuantumCircuit(sub_mps.num_sites)
            if len(qidxs) == 0:
                qidxs.append(list(range(sub_mps.num_sites)))
            else:
                qidxs.append(
                    list(
                        range(qidxs[-1][-1] + 1, qidxs[-1][-1] + 1 + sub_mps.num_sites)
                    )
                )
            for uni_idx in range(sub_mps.num_sites - 2):
                uni = unitaries[uni_idx]
                uni = uni[[0, 2, 1, 3], :]
                gate = UnitaryGate(uni)
                qc.append(gate, [uni_idx, uni_idx + 1])
            penultimate_uni = unitaries[-2]
            penultimate_uni = penultimate_uni[[0, 2, 1, 3], :]
            final_uni = unitaries[-1]
            final_uni_2q = np.kron(np.eye(2), final_uni)
            final_uni_2q = final_uni_2q[[0, 2, 1, 3], :]
            final_uni_2q = final_uni_2q[:, [0, 2, 1, 3]]
            total_uni = final_uni_2q @ penultimate_uni
            last_gate = UnitaryGate(total_uni)
            qc.append(last_gate, [sub_mps.num_sites - 2, sub_mps.num_sites - 1])
            qcs.append(qc)

        final_qc = QuantumCircuit(mps.num_sites)
        for qc_idx in range(len(qcs)):
            final_qc.compose(qcs[qc_idx], qidxs[qc_idx], inplace=True)

        return final_qc

    def disentangle_mps(
        self, mps: MatrixProductState, qc_layer: QuantumCircuit
    ) -> MatrixProductState:
        """Update the current MPS by diesntangling with a circuit layer"""
        sim = CircuitSimulator(qc_layer.inverse(), mps)
        out = sim.run()
        return out

    def calculate_fidelity(self, circ) -> float:
        """Calculate current fidelity"""
        state = copy.deepcopy(self.mps)
        sim = CircuitSimulator(circ)
        output = sim.run()
        fid = state_uhlmann_fidelity(output, state)
        return fid

    def run(self) -> QuantumCircuit:
        """Run the analytic decomposition"""
        while (
            self.num_layers < self.max_layers and self.fidelity < self.target_fidelity
        ):
            original_mps = copy.deepcopy(self.mps)
            disentangled_mps = self.disentangle_mps(original_mps, self.qc)
            bond_dim_2_mps = self.compress_to_bond_dim_2(disentangled_mps)
            qc_layer = self.bond_dim_2_to_qc_exact(bond_dim_2_mps)
            temp_circ = self.qc.compose(qc_layer, front=True)
            new_fidelity = self.calculate_fidelity(temp_circ)
            # if new_fidelity < self.fidelity:
            #     break
            self.qc = temp_circ
            self.fidelity = new_fidelity
            self.num_layers += 1
        return self.qc


class MPStoCircuit:
    def __init__(
        self, mps: MatrixProductState, max_layers: int, target_fidelity: float
    ):
        self.mps = mps
        self.max_layers = max_layers
        self.target_fidelity = target_fidelity
        self.num_layers = 0
        self.fidelity = 0.0
        self.qc = QuantumCircuit(mps.num_sites)
        self.current_mps = copy.deepcopy(mps)

    def calculate_fidelity(self, circ) -> float:
        """Calculate current fidelity"""
        state = copy.deepcopy(self.mps)
        sim = CircuitSimulator(circ)
        output = sim.run()
        fid = state_uhlmann_fidelity(output, state)
        return fid

    def disentangle_mps(self, qc_layer: QuantumCircuit) -> None:
        """Update the current MPS by diesntangling with a circuit layer"""
        current_mps = copy.deepcopy(self.current_mps)
        sim = CircuitSimulator(qc_layer.inverse(), current_mps)
        self.current_mps = sim.run()
        return

    def run(self, num_optimiser_sweeps: int = 1) -> QuantumCircuit:
        while (
            self.num_layers < self.max_layers and self.fidelity < self.target_fidelity
        ):
            qc_layer = MPSAnalyticDecomposition(self.current_mps, 1, 1.0).run()
            # qc_layer = MPSOptimiser(qc_layer, self.current_mps).run(num_optimiser_sweeps)
            temp_circ = self.qc.compose(qc_layer, front=True)
            new_fidelity = self.calculate_fidelity(temp_circ)
            if new_fidelity < self.fidelity:
                break
            self.qc = temp_circ
            self.disentangle_mps(qc_layer)
            self.fidelity = new_fidelity
            self.num_layers += 1
        # self.qc = MPSOptimiser(self.qc, self.mps).run(num_optimiser_sweeps)
        return self.qc



import copy
from timeit import default_timer

import numpy as np
from qiskit import QuantumCircuit

from ...quantum_algorithms.hamiltonian_simulation.qdrift import (
    QDriftSimulation,
)
from ...quantum_algorithms.hamiltonian_simulation.trotterisation import (
    TrotterSimulation,
)
from ..backend.base import QuantumBackend
from ..backend.tn_backend import TNQuantumBackend
from ..result import Result
from ..utils import add_controls
from .qsci import QSCI


class ControlledTimeEvolvedQSCI(QSCI):
    def __init__(
        self,
        hamiltonian: dict,
        hf_state: str,
        backend: QuantumBackend | None = None,
        duration: float = np.pi,
        num_circuits: int = 5,
        qdrift: bool = True,
        **kwargs,
    ) -> "QSCI":
        """
        Constructor for QSCI class.
        """
        self.duration = duration
        self.num_circuits = num_circuits
        self.qdrift_config = kwargs
        self.qdrift = qdrift
        hamiltonian = self.sanitize_dict(hamiltonian)
        super().__init__(hamiltonian, hf_state, backend)

    @property
    def circuit(self) -> QuantumCircuit:
        return self._circuit

    def sanitize_dict(self, d: dict[str, complex | float]) -> dict[str, float]:
        return {
            k: float(v.real) if isinstance(v, complex) else float(v)
            for k, v in d.items()
        }

    def perform_controlled_time_evolution(self, duration: float) -> QuantumCircuit:
        """Add time evolution to the circuit"""
        if duration == 0.0:
            ref = copy.deepcopy(self.circuit)
            return ref
        sim = TrotterSimulation(self.hamiltonian, duration=duration)
        sim_circ = sim.circuit
        controlled_sim_circ = QuantumCircuit(sim_circ.num_qubits + 1)
        controlled_sim_circ.compose(
            sim_circ, qubits=range(1, sim_circ.num_qubits + 1), inplace=True
        )
        controlled_sim_circ = add_controls(controlled_sim_circ, [0])
        circ_copy = copy.deepcopy(self.circuit)
        ref = QuantumCircuit(circ_copy.num_qubits + 1)
        ref.compose(circ_copy, qubits=range(1, circ_copy.num_qubits + 1), inplace=True)
        ref.compose(controlled_sim_circ, inplace=True)
        return ref

    def perform_controlled_time_evolution_qdrift(
        self, duration: float, error: float | None = None
    ) -> QuantumCircuit:
        """Add qdrift time evolution to the circuit"""
        if duration == 0.0:
            ref = copy.deepcopy(self.circuit)
            return ref
        sim = QDriftSimulation(self.hamiltonian, duration=duration, error=error)
        sim_circ = sim.circuit
        controlled_sim_circ = QuantumCircuit(sim_circ.num_qubits + 1)
        controlled_sim_circ.compose(
            sim_circ, qubits=range(1, sim_circ.num_qubits + 1), inplace=True
        )
        controlled_sim_circ = add_controls(controlled_sim_circ, [0])
        circ_copy = copy.deepcopy(self.circuit)
        ref = QuantumCircuit(circ_copy.num_qubits + 1)
        ref.compose(circ_copy, qubits=range(1, circ_copy.num_qubits + 1), inplace=True)
        ref.compose(controlled_sim_circ, inplace=True)
        return ref

    def post_selection(self, counts: dict[str, int]) -> dict[str, int]:
        """Post select counts based on the ancilla output"""
        if len(list(counts.keys())[0]) == self.hamiltonian_mpo.num_sites:
            return counts
        new_counts = {}
        for b, count in counts.items():
            if b[0] == "0":
                new_counts[b[1:]] = new_counts.get(b[1:], 0) + count
        return new_counts

    def get_counts(
        self, duration: float, num_circuits: int, shots: int
    ) -> dict[str, int]:
        """Get counts using Trotterisation"""
        duration_per_circuit = duration / (num_circuits - 1)
        counts = {}
        for idx in range(num_circuits):
            qc = self.perform_controlled_time_evolution(idx * duration_per_circuit)
            subcounts = self.backend.run(qc, shots=shots)
            for b, count in subcounts.items():
                if idx == 0:
                    b = "0" + b
                counts[b] = counts.get(b, 0) + count
        return counts

    def get_counts_qdrift(
        self, duration: float, num_circuits: int, shots: int
    ) -> dict[str, int]:
        """Get counts using qDRIFT"""
        duration_per_circuit = duration / (num_circuits - 1)
        counts = {}
        for idx in range(num_circuits):
            shots_per_circuit = int(shots / self.qdrift_config["num_qdrift_circuits"])
            for _ in range(self.qdrift_config["num_qdrift_circuits"]):
                qc = self.perform_controlled_time_evolution_qdrift(
                    idx * duration_per_circuit, error=self.qdrift_config["error"]
                )
                subcounts = self.backend.run(qc, shots=shots_per_circuit)
                for b, count in subcounts.items():
                    if idx == 0:
                        b = "0" + b
                    counts[b] = counts.get(b, 0) + count
        return counts

    def run(
        self, num_shots: int, subspace_size: int, num_iterations: int = 1
    ) -> Result:
        """Run the full algorithm pipeline. Returns result object or final value."""
        start_time = default_timer()
        for idx in range(num_iterations):
            print("Starting iteration", idx + 1)
            if self.hfs:
                self.state, _ = self.hf_suppression(self.state)
            self._circuit = self.prepare_state(self.state)
            if self.qdrift:
                counts = self.get_counts_qdrift(
                    self.duration, self.num_circuits, num_shots
                )
            else:
                counts = self.get_counts(self.duration, self.num_circuits, num_shots)
            ps_counts = self.post_selection(counts)
            cr_counts = self.configuration_recovery(ps_counts, self.num_electrons)
            samples = self.gather_samples(cr_counts, subspace_size)
            if self.hfs and self.hf_state not in samples:
                samples += self.hf_state
            print(f"Collected {len(samples)} unique samples")
            if len(samples) <= 500:
                projected_ham = self.project_hamiltonian(samples)
                self.energy, groundstate_vec = self.exact_diagonalisation(projected_ham)
            else:
                self.energy, groundstate_vec = self.linear_operator_diagonalisation(
                    samples
                )
            print(f"New groundtate found with energy {self.energy}")
            self.state = self.reconstruct_mps(samples, groundstate_vec)
            print("Finished iteration", idx + 1)
        end_time = default_timer()

        metadata = {
            "algorithm_name": "Time Evolved QSCI",
            "qdrift": self.qdrift,
            "num_shots": num_shots,
            "num_iterations": num_iterations,
            "max_subspace_size": subspace_size,
            "actual_subspace_size": len(samples),
            "total_runtime": end_time - start_time,
        }
        if self.backend is not None:
            metadata["backend_name"] = self.backend.name
            metadata["backend_coupling_map"] = self.backend.coupling_map
            metadata["backend_basis_gates"] = self.backend.basis_gates
            metadata["backend_num_qubits"] = self.backend.num_qubits
        if self.qdrift:
            metadata.update(self.qdrift_config)

        result = self.energy

        result = Result(
            result=result,
            measurements=counts,
            parameters=None,
            metadata=metadata,
        )
        return result

    def set_backend(self, backend: QuantumBackend | None) -> None:
        """Attach a QuantumBackend instance for execution."""
        if backend is None:
            backend = TNQuantumBackend()
        self.backend = backend
        return



from collections.abc import Callable
from timeit import default_timer
from typing import Union

import matplotlib.pyplot as plt
from numpy import ndarray
from qiskit import QuantumCircuit
from qiskit.primitives import Estimator
from qiskit.quantum_info import SparsePauliOp
from qiskit_algorithms import VQE
from qiskit_algorithms.optimizers import Optimizer

from ..backend.base import QuantumBackend
from ..base import QuantumAlgorithm
from ..result import Result
from .ansatz_circuits import (
    hea_ansatz,
    number_preserving_ansatz,
    pauli_two_design_ansatz,
)
from .classical_optimisers import (
    adam_optimiser,
    bfgs_optimiser,
    cobyla_optimiser,
    qnspsa_optimiser,
)


class VQEAlgorithm(QuantumAlgorithm):
    def __init__(
        self,
        hamiltonian: dict,
        num_electrons: int | None = None,
        max_iterations_vqe: int = 1e3,
        convergence_threshold: float = 1e-6,
        initial_points: ndarray = None,
        ansatz: Union[QuantumCircuit, str] | None = None,
        optimiser: Union[Optimizer, str] | None = None,
        estimator: Estimator | None = None,
        backend: QuantumBackend | None = None,
    ) -> None:
        """
        Constructor

        Args:
            hamiltonian: a qubit Hamiltonian given as a dict
            num_electrons: Required if using number preserving ansatz
            max_iterations_vqe: Maximum number of iterations
            convergence_threshold: Convergence threshold for early exit
            initial_points: Initial starting points
            ansatz: Ansatz given as a QuantumCircuit or a string (number_preserving_ansatz, hardware_efficient_ansatz, pauli_two_design_ansatz)
            optimiser: Optimiser given as an Optimiser or a string (QNSPSA, ADAM, BFGS, COBYLA)
            estimator: Estimator given as an Estimator or None
            backend: QuantumBackend, currently only Qiskit simulation
        """

        if isinstance(hamiltonian, dict):
            self.num_qubits = len(list(hamiltonian.keys())[0])
        else:
            self.num_qubits = len(hamiltonian[0])
        self.hamiltonian = hamiltonian
        self.num_electrons = num_electrons
        self.optimisation_index = 0
        self.max_iterations_vqe = max_iterations_vqe
        self.convergence_threshold = convergence_threshold
        self.initial_points = initial_points

        self.clear_optimisation_dict()
        self.clear_warm_starting_dict()

        self.set_backend(backend=backend)

        self.set_ansatz(ansatz=ansatz)
        self.set_optimiser(optimiser=optimiser)
        self.set_estimator(estimator=estimator)

        self.set_callback()
        self.driver = self.vqe_driver()

    @property
    def circuit(self) -> QuantumCircuit:
        return self.ansatz

    def set_estimator(self, estimator: Estimator | None = None) -> None:
        """Set the Estimator

        Args:
            estimator: The Estimator, optional, defaults to Estimator()
        """
        if not estimator:
            self.estimator = Estimator()
        else:
            self.estimator = estimator
        return

    def set_ansatz(self, ansatz: QuantumCircuit | str | None = None) -> None:
        """Set ansatz circuit

        Args:
            ansatz: QuantumCircuit, string identifier or None, defaults to number_preserving_ansatz
        """
        if not ansatz or ansatz == "number_preserving_ansatz":
            qc1 = QuantumCircuit(self.num_qubits)
            qc1.x(range(self.num_electrons))
            qc2 = number_preserving_ansatz(self.num_qubits, 3, "linear")
            qc = qc1.compose(qc2)
            self.ansatz = qc
        elif ansatz == "hardware_efficient_ansatz":
            qc = hea_ansatz(self.num_qubits)
            self.ansatz = qc
        elif ansatz == "pauli_two_design_ansatz":
            qc = pauli_two_design_ansatz(self.num_qubits)
            self.ansatz = qc
        elif isinstance(ansatz, str):
            print(
                "Ansatz string not recognised. Must be one of 'hea_ansatz', 'number_preserving_ansatz', 'pauli_two_design_ansatz', 'uccsd_ansatz'"
            )
        else:
            assert isinstance(ansatz, QuantumCircuit)
            self.ansatz = ansatz
        return

    def set_optimiser(self, optimiser: Optimizer | str | None = None) -> None:
        """Set optimiser

        Args:
            optimiser: Optimiser or string identifier or None, defaults to QNSPSA
        """
        self.optimisation_index = 0
        if not optimiser or optimiser == "QNSPSA":
            self.optimiser = qnspsa_optimiser(
                self.ansatz,
                self.max_iterations_vqe,
                opt_dict=self.optimisation_dict,
                index=self.optimisation_index,
            )
        elif optimiser == "ADAM":
            self.optimiser = adam_optimiser(
                self.max_iterations_vqe,
                opt_dict=self.optimisation_dict,
                index=self.optimisation_index,
            )
        elif optimiser == "BFGS":
            self.optimiser = bfgs_optimiser(
                self.max_iterations_vqe,
                opt_dict=self.optimisation_dict,
                index=self.optimisation_index,
            )
        elif optimiser == "COBYLA":
            self.optimiser = cobyla_optimiser(
                self.max_iterations_vqe,
                self.convergence_threshold,
                opt_dict=self.optimisation_dict,
                index=self.optimisation_index,
            )
        elif isinstance(optimiser, str):
            print(
                "Optimiser string not recognised. Must be one of 'ADAM', 'COBYLA', 'QNSPSA', 'BFGS'"
            )
        else:
            assert isinstance(optimiser, Optimizer)
            self.optimiser = optimiser
        return

    def set_callback(self, callback: Callable | None = None) -> None:
        """Set callback

        Args:
            callback: Function or None
        """
        if not callback:

            def default_callback(i, a, f, _):
                # print(i, a, f)
                self.optimisation_dict["optimisation_number"].append(i)
                self.optimisation_dict["optimisation_parameters"].append(a)
                self.optimisation_dict["optimisation_value"].append(f)
                return

            self.callback = default_callback
        else:
            self.callback = callback
        return

    def clear_optimisation_dict(self) -> None:
        """Reset optimisation dictionary"""
        self.optimisation_dict = {
            "optimisation_number": [],
            "optimisation_parameters": [],
            "optimisation_value": [],
        }
        return

    def clear_warm_starting_dict(self) -> None:
        """Reset warm starting dictionary"""
        self.warm_starting_dict = {
            "optimisation_number": [],
            "optimisation_parameters": [],
            "optimisation_value": [],
        }
        return

    def vqe_driver(self) -> VQE:
        """Set the VQE driver"""
        driver = VQE(
            self.estimator,
            self.ansatz,
            self.optimiser,
            initial_point=self.initial_points,
            callback=self.callback,
        )
        return driver

    def plot_convergence(self) -> None:
        """Plot convergence data"""
        plt.plot(
            self.optimisation_dict["optimisation_number"],
            self.optimisation_dict["optimisation_value"],
        )
        plt.show()
        return

    def run(self) -> Result:
        """Run the full algorithm pipeline. Returns result object."""
        pauli_list = list(self.hamiltonian.keys())
        coeffs = list(self.hamiltonian.values())
        observable = SparsePauliOp(pauli_list, coeffs)
        self.clear_optimisation_dict()
        start_time = default_timer()
        result = self.driver.compute_minimum_eigenvalue(observable)
        end_time = default_timer()
        self.minimum_eigenvalue = result.eigenvalue
        metadata = {
            "algorithm_name": "VQE",
            "num_parameters": len(
                self.optimisation_dict["optimisation_parameters"][-1]
            ),
            "num_iterations": self.optimisation_dict["optimisation_number"][-1],
            "total_runtime": end_time - start_time,
        }
        if self.backend is not None:
            metadata["backend_name"] = self.backend.name
            metadata["backend_coupling_map"] = self.backend.coupling_map
            metadata["backend_basis_gates"] = self.backend.basis_gates
            metadata["backend_num_qubits"] = self.backend.num_qubits
        result = Result(
            result=self.minimum_eigenvalue,
            measurements=None,
            parameters=self.optimisation_dict["optimisation_parameters"][-1],
            metadata=metadata,
        )
        return result

    def set_backend(self, backend: QuantumBackend | None = None) -> None:
        """Attach a QuantumBackend instance for execution.

        Args:
            backend: Currently only supports Qiskit backends
        """
        self.backend = backend
        return



import copy

from qiskit import QuantumCircuit, transpile

from ...circuit_simulator import CircuitSimulator
from ...mps import MatrixProductState
from .base import QuantumBackend


class TNQuantumBackend(QuantumBackend):
    """
    Backend using TN4QA's CircuitSimulator for circuit execution
    """

    def __init__(self) -> None:
        """Constructor"""
        self._name = "tn4qa_circuit_simulator"
        self._coupling_map = None
        self._basis_gates = None
        self._num_qubits = None

    @property
    def name(self) -> str:
        return self._name

    @property
    def coupling_map(self) -> list[tuple[int, int]]:
        return self._coupling_map

    @property
    def num_qubits(self) -> int:
        return self._num_qubits

    @property
    def basis_gates(self) -> list[str]:
        return self._basis_gates

    def run(
        self,
        circuit: QuantumCircuit,
        shots: int,
        max_bond: int | None = None,
        input_state: MatrixProductState | None = None,
    ) -> dict[str, int]:
        """Execute the circuit

        Args:
            circuit: The QuantumCircuit object to run
            shots: If provided will sample from the resulting state
            max_bond: The maximum bond dimension allowed

        Returns:
            Measurement results {bitstring : count}
        """
        qc = copy.deepcopy(circuit)
        qc = transpile(qc, basis_gates=["u", "cx"])
        sim = CircuitSimulator(qc, input_state=input_state)
        output = sim.run(max_bond_dimension=max_bond, samples=shots)
        return output

    def parse_openqasm(self, filename: str) -> QuantumCircuit:
        """Parse an OpenQASM input circuit

        Args:
            filename: The filename of the OpenQASM input

        Returns:
            A qiskit QuantumCircuit object
        """
        qc = QuantumCircuit.from_qasm_file(filename)
        return qc

    def get_device_info(self) -> dict:
        """Return a dictionary describing the backend."""
        return {"backend_name": self.name}



from abc import ABC, abstractmethod


class QuantumBackend(ABC):
    """
    A class for quantum backends (simulated and real)
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """Return the backend name"""
        pass

    @property
    @abstractmethod
    def coupling_map(self) -> str:
        """Return the backend coupling map"""
        pass

    @property
    @abstractmethod
    def num_qubits(self) -> str:
        """Return the backend num qubits"""
        pass

    @property
    @abstractmethod
    def basis_gates(self) -> str:
        """Return the backend basis gates"""
        pass

    @abstractmethod
    def run(self, circuit, shots: int, **kwargs) -> dict[str, int]:
        """Execute the circuit"""
        pass

    @abstractmethod
    def parse_openqasm(self, filename: str):
        """Parse an OpenQASM input circuit"""
        pass

    @abstractmethod
    def get_device_info(self) -> dict:
        """Return a dictionary describing the backend device (connectivity, noise, etc)."""
        pass



from timeit import default_timer

import numpy as np
from qiskit import QuantumCircuit

from ..backend.base import QuantumBackend
from ..backend.tn_backend import TNQuantumBackend
from ..base import QuantumAlgorithm
from ..result import Result
from ..utils import calculate_exp_val, exp_pauli_string_to_circ


class TrotterSimulation(QuantumAlgorithm):
    """
    Perform Hamiltonian simulation by Trotterisation
    """

    def __init__(
        self,
        hamiltonian: dict[str, float],
        duration: float,
        error: float = 0.05,
        num_steps: int | None = None,
        backend: QuantumBackend | None = None,
    ) -> "TrotterSimulation":
        """
        Constructor for Trotter simulation class.

        Args:
            hamiltonian: The qubit Hamiltonian
            duration: The time to simulate evolution for
            num_steps: The number of Trotter steps, defaults to a sensible value
        """
        self.hamiltonian = hamiltonian

        norm = np.sum([np.abs(x) for x in hamiltonian.values()])
        self.duration = duration
        if not num_steps:
            self.num_steps = int(np.ceil((duration**2 * norm**2) / error))
        else:
            self.num_steps = num_steps
        self.delta_t = duration / self.num_steps

        pauli_strings = list(hamiltonian.keys())

        num_qubits = len(pauli_strings[0])
        qc = QuantumCircuit(num_qubits)

        for _ in range(self.num_steps):
            for p in pauli_strings:
                temp_qc = exp_pauli_string_to_circ(p, self.delta_t * hamiltonian[p])
                qc.compose(temp_qc, inplace=True)

        self._circuit = qc
        self.set_backend(backend)

    @property
    def circuit(self) -> QuantumCircuit:
        return self._circuit

    def run(self, num_shots: int = 1024, observable: dict | None = None) -> Result:
        """Run the full algorithm pipeline. Returns result object or final value."""
        start_time = default_timer()
        if observable is None:
            counts = self.backend.run(self.circuit, shots=num_shots)
            result = None
        else:
            result = calculate_exp_val(
                self.circuit, observable, self.backend, num_shots
            )
            counts = None

        end_time = default_timer()

        metadata = {
            "algorithm_name": "Trotterisation",
            "num_shots": num_shots,
            "total_runtime": end_time - start_time,
        }
        if self.backend is not None:
            metadata["backend_name"] = self.backend.name
            metadata["backend_coupling_map"] = self.backend.coupling_map
            metadata["backend_basis_gates"] = self.backend.basis_gates
            metadata["backend_num_qubits"] = self.backend.num_qubits
        result = Result(
            result=result,
            measurements=counts,
            parameters=None,
            metadata=metadata,
        )
        return result

    def set_backend(self, backend: QuantumBackend | None = None) -> None:
        """Attach a QuantumBackend instance for execution."""
        if backend is None:
            backend = TNQuantumBackend()
        self.backend = backend
        return



import copy

import numpy as np
import sparse
from numpy.linalg import svd
from qiskit import QuantumCircuit
from qiskit.circuit import CircuitInstruction
from qiskit.circuit.library import UnitaryGate
from qiskit.quantum_info import Operator
from scipy.sparse.linalg import svds
from sparse import COO

from tn4qa.mpo import MatrixProductOperator
from tn4qa.mps import MatrixProductState


class CircuitSimulator:
    """
    A class to simulate quantum circuits built using Qiskit
    """

    def __init__(
        self, circuit: QuantumCircuit, input_state: MatrixProductState | None = None
    ) -> None:
        """
        Class constructor.

        Args:
            circuit: The Qiskit QuantumCircuit object
        """
        self.circuit = circuit
        self.num_qubits = circuit.num_qubits
        self.set_input_state(input_state)
        self.current_state = copy.deepcopy(self.input_state)
        self.output_state = None
        self.site_mapping = {
            str(idx): str(idx) for idx in range(1, self.num_qubits + 1)
        }  # {physical_site:logical_site}
        self.mpo = MatrixProductOperator.identity_mpo(self.num_qubits)

    def set_input_state(self, input_state: MatrixProductState | None) -> None:
        """
        Set the input state to the circuit

        Args:
            input_state: The input state, defaults to the all zero state
        """
        if not input_state:
            input_state = MatrixProductState.all_zero_mps(self.num_qubits)

        self.input_state = input_state
        self.input_state.set_default_indices()

    def apply_one_qubit_gate(self, data: COO, site: int) -> None:
        """
        Apply a one-qubit gate in place

        Args:
            data: The one-qubit matrix
            site: Where to apply the gate to
        """
        site_loc = int(self.site_mapping[str(site)])
        tensor = self.current_state.tensors[site_loc - 1]
        if site_loc == 1 or site_loc == self.num_qubits:
            contraction = "ij,kj->ik"
        else:
            contraction = "ijk,lk->ijl"
        self.current_state.tensors[site_loc - 1].data = sparse.einsum(
            contraction, tensor.data, data
        )
        return

    def move_site_to_location(self, site_source: int, site_destination: int) -> None:
        """
        Move a site to a different location

        Args:
            site_source: The starting location
            site_destination: The final location
        """
        if site_source < site_destination:
            for site in range(site_source, site_destination):
                self.current_state.swap_neighbouring_sites(site)
                reverse_mapping = {v: k for k, v in self.site_mapping.items()}
                (
                    self.site_mapping[reverse_mapping[str(site)]],
                    self.site_mapping[reverse_mapping[str(site + 1)]],
                ) = (
                    str(site + 1),
                    str(site),
                )
        else:
            for site in range(site_source, site_destination, -1):
                self.current_state.swap_neighbouring_sites(site - 1)
                reverse_mapping = {v: k for k, v in self.site_mapping.items()}
                (
                    self.site_mapping[reverse_mapping[str(site - 1)]],
                    self.site_mapping[reverse_mapping[str(site)]],
                ) = (
                    str(site),
                    str(site - 1),
                )
        return

    def apply_two_qubit_gate(
        self,
        data: COO,
        sites: list[int],
        max_bond: int | None = None,
        tol: float = 1e-12,
    ) -> None:
        """
        Apply a two qubit gate in place

        Args:
            data: The two-qubit matrix
            sites: The sites to apply it to
            max_bond: The maximum allowed bond dimension
        """
        sites_locs = [int(self.site_mapping[str(site)]) for site in sites]
        site0, site1 = sites_locs[0], sites_locs[1]
        data = sparse.reshape(data, (2, 2, 2, 2))

        if self.num_qubits == 2:
            data = sparse.moveaxis(data, [0, 1, 2, 3], [1, 0, 3, 2])
            if site1 < site0:
                data = sparse.moveaxis(data, [0, 1, 2, 3], [1, 0, 3, 2])
            data = sparse.reshape(data, (4, 4))
            gate = UnitaryGate(data.todense())
            qc = QuantumCircuit(2)
            qc.append(gate, [site0 - 1, site1 - 1])
            gate_mpo = MatrixProductOperator.from_qiskit_gate(qc.data[0])
            self.current_state = self.current_state.apply_mpo(gate_mpo)
            return

        if site1 < site0:
            data = sparse.moveaxis(data, [0, 1, 2, 3], [1, 0, 3, 2])
            if site1 == site0 - 1:
                pass
            else:
                self.move_site_to_location(site1, site0 - 1)
            tensor0 = self.current_state.tensors[site0 - 2]
            tensor1 = self.current_state.tensors[site0 - 1]
            if site0 - 1 == 1:
                contraction = "ij,ikl,mnjl->knm"
                output_shape = (tensor1.dimensions[1], 2, 2)
                mat_shape = (2 * tensor1.dimensions[1], 2)
            elif site0 == self.num_qubits:
                contraction = "hij,ik,lmjk->mhl"
                output_shape = (2, tensor0.dimensions[0], 2)
                mat_shape = (2, tensor0.dimensions[0] * 2)
            else:
                contraction = "hij,ikl,mnjl->knhm"
                output_shape = (tensor1.dimensions[1], 2, tensor0.dimensions[0], 2)
                mat_shape = (tensor1.dimensions[1] * 2, tensor0.dimensions[0] * 2)
        else:
            if site1 == site0 + 1:
                pass
            else:
                self.move_site_to_location(site1, site0 + 1)
            tensor0 = self.current_state.tensors[site0 - 1]
            tensor1 = self.current_state.tensors[site0]
            if site0 == 1:
                contraction = "ij,ikl,mnjl->knm"
                output_shape = (tensor1.dimensions[1], 2, 2)
                mat_shape = (2 * tensor1.dimensions[1], 2)
            elif site0 + 1 == self.num_qubits:
                contraction = "hij,ik,lmjk->mhl"
                output_shape = (2, tensor0.dimensions[0], 2)
                mat_shape = (2, tensor0.dimensions[0] * 2)
            else:
                contraction = "hij,ikl,mnjl->knhm"
                output_shape = (tensor1.dimensions[1], 2, tensor0.dimensions[0], 2)
                mat_shape = (tensor1.dimensions[1] * 2, tensor0.dimensions[0] * 2)

        output_data = sparse.einsum(contraction, tensor0.data, tensor1.data, data)
        output_data = np.reshape(output_data, mat_shape)

        if max_bond:
            bond_dim = min([max_bond, mat_shape[0], mat_shape[1]])
        else:
            bond_dim = min([mat_shape[0], mat_shape[1]])

        if bond_dim >= min([mat_shape[0], mat_shape[1]]) - 1:
            u, s, vh = svd(output_data.todense(), full_matrices=False)
        else:
            u, s, vh = svds(output_data, k=bond_dim)

        s = s[s > 1e-14]
        sq = s**2
        cumulative = np.cumsum(sq[::-1])[::-1]
        keep_dim = len(s)
        for k in range(len(s)):
            if cumulative[k] < tol**2:
                keep_dim = k + 1
                break
        keep_dim = min(keep_dim, bond_dim)

        new_data0 = sparse.COO.from_numpy(vh[:keep_dim, :])
        new_data1 = sparse.COO.from_numpy(u[:, :keep_dim] * s[:keep_dim])

        if site1 < site0:
            if site0 - 1 == 1:
                new_data0 = sparse.reshape(new_data0, (keep_dim,) + (output_shape[2],))
                new_data1 = sparse.reshape(new_data1, output_shape[:2] + (keep_dim,))
                new_data1 = sparse.moveaxis(new_data1, [2], [0])
            elif site0 == self.num_qubits:
                new_data0 = sparse.reshape(new_data0, (keep_dim,) + output_shape[-2:])
                new_data0 = sparse.moveaxis(new_data0, [0], [1])
                new_data1 = sparse.reshape(new_data1, (output_shape[0],) + (keep_dim,))
                new_data1 = sparse.moveaxis(new_data1, [0], [1])
            else:
                new_data0 = sparse.reshape(new_data0, (keep_dim,) + output_shape[-2:])
                new_data0 = sparse.moveaxis(new_data0, [0], [1])
                new_data1 = sparse.reshape(new_data1, output_shape[:2] + (keep_dim,))
                new_data1 = sparse.moveaxis(new_data1, [2], [0])
            self.current_state.tensors[site0 - 2].data = new_data0
            self.current_state.tensors[
                site0 - 2
            ].dimensions = self.current_state.tensors[site0 - 2].data.shape
            self.current_state.tensors[site0 - 1].data = new_data1
            self.current_state.tensors[
                site0 - 1
            ].dimensions = self.current_state.tensors[site0 - 1].data.shape
            self.current_state.bond_dims = [
                t.dimensions[0] for t in self.current_state.tensors[1:]
            ]
            self.current_state.bond_dimension = max(self.current_state.bond_dims)
        else:
            if site0 == 1:
                new_data0 = sparse.reshape(new_data0, (keep_dim,) + (output_shape[2],))
                new_data1 = sparse.reshape(new_data1, output_shape[:2] + (keep_dim,))
                new_data1 = sparse.moveaxis(new_data1, [2], [0])
            elif site0 + 1 == self.num_qubits:
                new_data0 = sparse.reshape(new_data0, (keep_dim,) + output_shape[-2:])
                new_data0 = sparse.moveaxis(new_data0, [0], [1])
                new_data1 = sparse.reshape(new_data1, (output_shape[0],) + (keep_dim,))
                new_data1 = sparse.moveaxis(new_data1, [0], [1])
            else:
                new_data0 = sparse.reshape(new_data0, (keep_dim,) + output_shape[-2:])
                new_data0 = sparse.moveaxis(new_data0, [0], [1])
                new_data1 = sparse.reshape(new_data1, output_shape[:2] + (keep_dim,))
                new_data1 = sparse.moveaxis(new_data1, [2], [0])
            self.current_state.tensors[site0 - 1].data = new_data0
            self.current_state.tensors[
                site0 - 1
            ].dimensions = self.current_state.tensors[site0 - 1].data.shape
            self.current_state.tensors[site0].data = new_data1
            self.current_state.tensors[site0].dimensions = self.current_state.tensors[
                site0
            ].data.shape
            self.current_state.bond_dims = [
                t.dimensions[0] for t in self.current_state.tensors[1:]
            ]
            self.current_state.bond_dimension = max(self.current_state.bond_dims)
        return

    def apply_general_gate(
        self,
        inst: CircuitInstruction,  # type: ignore
        max_bond: int | None = None,
    ) -> None:  # type: ignore
        """
        Apply a gate with no better option

        Args:
            inst: The circuit instruction
        """
        qidxs = [inst.qubits[i]._index + 1 for i in range(inst.operation.num_qubits)]
        qidxs = [int(self.site_mapping[str(qidx)]) for qidx in qidxs]
        mpo = MatrixProductOperator.from_qiskit_gate(inst)
        self.current_state = self.current_state.apply_sub_mpo(mpo, qidxs, max_bond)
        return

    def restore_ordering(self) -> None:
        """
        Restore correct site ordering
        """
        reversed_mapping = {v: k for k, v in self.site_mapping.items()}
        target_site_ordering = [
            int(reversed_mapping[str(site)]) for site in range(1, self.num_qubits + 1)
        ]
        self.current_state.reorder_sites(target_site_ordering)
        self.site_mapping = {
            str(idx): str(idx) for idx in range(1, self.num_qubits + 1)
        }
        return

    def run(
        self, max_bond_dimension: int | None = None, samples: int | None = None
    ) -> MatrixProductState | dict[str, int]:
        """
        Execute the quantum circuit

        Args:
            max_bond_dimension: The maximum allowed bond dimension
            samples: If provided will return this number of bitstring samples from the output state
        """
        for inst in self.circuit.data:
            qidxs = [
                inst.qubits[i]._index + 1 for i in range(inst.operation.num_qubits)
            ]
            data = COO.from_numpy(Operator(inst.operation).reverse_qargs().data)
            if len(qidxs) == 1:
                self.apply_one_qubit_gate(data, qidxs[0])
            elif len(qidxs) == 2:
                self.apply_two_qubit_gate(data, qidxs, max_bond_dimension)
            else:
                self.apply_general_gate(inst, max_bond_dimension)
        self.restore_ordering()
        self.output_state = self.current_state

        if samples:
            sample_dict = self.output_state.sample_bitstrings(samples)
            return sample_dict

        return self.output_state

    def get_operator_mpo(
        self, after_gate: int | None = None, max_bond: int | None = None
    ) -> MatrixProductOperator:
        """
        Build the MPO representing the quantum circuit

        Args:
            after_gate: Builds the MPO representing the circuit up to after the given gate number. Defaults to full circuit
            max_bond: Maximum allowed bond dimension

        Returns:
            An MPO
        """
        if not after_gate:
            qc_data = self.circuit.data
        else:
            qc_data = self.circuit.data[:after_gate]

        qc_after_gate = QuantumCircuit(self.num_qubits)
        for inst in qc_data:
            qc_after_gate.append(inst)
        mpo = MatrixProductOperator.from_qiskit_circuit(qc_after_gate, max_bond)
        return mpo



from timeit import default_timer
from typing import TypeAlias, Union

import numpy as np
from numpy import ndarray
from qiskit import QuantumCircuit
from qiskit.circuit import CircuitInstruction, Operation
from qiskit.circuit.library import QFT
from sparse import SparseArray

from ..backend.base import QuantumBackend
from ..backend.tn_backend import TNQuantumBackend
from ..base import QuantumAlgorithm
from ..hamiltonian_simulation.qdrift import QDriftSimulation
from ..hamiltonian_simulation.trotterisation import TrotterSimulation
from ..result import Result
from ..utils import add_controls, count_qubits, to_quantum_circuit

TypeOptions: TypeAlias = Union[
    QuantumCircuit, Operation, CircuitInstruction, ndarray, SparseArray  # type: ignore
]  # type: ignore


class QPE(QuantumAlgorithm):
    def __init__(
        self,
        unitary: TypeOptions,
        state: TypeOptions,
        num_precision_bits: int,
        backend: QuantumBackend | None = None,
    ) -> "QPE":  # type: ignore
        """
        Constructor for QPE algorithm.

        Args:
            unitary: The unitary operation to estimate the phases for
            state: The input state to QPE
            num_precision_bits: The number of precision bits
            backend: The backend, defaults to TN4QA circuit simulator
        """
        self.num_state_qubits = count_qubits(state)
        self.precision_bits = num_precision_bits

        unitary_circ = to_quantum_circuit(unitary)
        state_circ = to_quantum_circuit(state)
        iqft = QFT(num_precision_bits, inverse=True).decompose()

        qc = QuantumCircuit(self.num_state_qubits + num_precision_bits)
        qc.h(range(num_precision_bits))
        for idx in range(num_precision_bits):
            temp_qc = QuantumCircuit(self.num_state_qubits + num_precision_bits)
            for _ in range(2**idx):
                temp_qc.compose(
                    unitary_circ,
                    qubits=range(
                        num_precision_bits, num_precision_bits + self.num_state_qubits
                    ),
                    inplace=True,
                )
            temp_qc = add_controls(temp_qc, [idx])
            qc.compose(temp_qc, inplace=True)
        qc.compose(
            state_circ,
            qubits=range(
                num_precision_bits, num_precision_bits + self.num_state_qubits
            ),
            inplace=True,
            front=True,
        )
        qc.compose(iqft, qubits=range(num_precision_bits), inplace=True)
        self._circuit = qc
        self.set_backend(backend=backend)
        self.from_ham = False
        self.t = None

    @property
    def circuit(self) -> QuantumCircuit:
        return self._circuit

    @classmethod
    def from_hamiltonian(
        cls,
        ham: dict,
        state: TypeOptions,
        num_precision_bits: int,
        backend: QuantumBackend | None = None,
        qdrift: bool = False,
    ) -> "QPE":
        """Set up QPE for a Hamiltonian input"""
        if qdrift:
            circ_builder = QDriftSimulation()
        else:
            circ_builder = TrotterSimulation()
        unitary = circ_builder.circuit
        qpe = cls(unitary, state, num_precision_bits, backend)
        qpe.from_ham = True
        norm = np.sum([np.abs(x) for x in ham.values()])
        qpe.t = np.pi / norm
        return qpe

    def measurement_to_phase(self, bitstring: str) -> float:
        """ "Estimate the phase value theta in e^{2*i*pi*theta} given a measured bitstring"""
        phase = 0
        bitstring = bitstring[: self.precision_bits][::-1]
        for b_idx in range(self.precision_bits):
            b = bitstring[b_idx]
            if b == "1":
                phase += 1 / (2 ** (b_idx + 1))
        return phase

    def collect_phase_distribution(self, measurements: dict) -> dict:
        """Estimate the phase values from a set of measurement results"""
        phase_freq_dict = {}
        for k, v in measurements.items():
            phase = self.measurement_to_phase(k)
            if phase in phase_freq_dict:
                phase_freq_dict[phase] += v
            else:
                phase_freq_dict[phase] = v
        return phase_freq_dict

    def get_most_likely_phase(self, phase_freq_dict: dict) -> float:
        """Get the most probable phase"""
        return max(phase_freq_dict, key=phase_freq_dict.get)

    def phase_to_energy(self, phase: float, t: float) -> float:
        """Assuming U = e^{-iHt} we can interpret the phase as an energy eigenvalue"""
        energy = (2 * np.pi * phase) / t
        return energy

    def run(self, num_shots: int = 1024) -> Result:
        """Run the full algorithm pipeline. Returns result object or final value.

        Args:
            num_shots: Number of shots to use per circuit
        """
        start_time = default_timer()
        counts = self.backend.run(self.circuit, shots=num_shots)
        self.phase_freq_dict = self.collect_phase_distribution(counts)
        self.phase = self.get_most_likely_phase(self.phase_freq_dict)
        if self.from_ham:
            self.energy = self.phase_to_energy(self.phase, self.t)
        end_time = default_timer()

        metadata = {
            "algorithm_name": "QPE",
            "num_shots": num_shots,
            "total_runtime": end_time - start_time,
            "phase_freq_dict": self.phase_freq_dict,
        }
        if self.backend is not None:
            metadata["backend_name"] = self.backend.name
            metadata["backend_coupling_map"] = self.backend.coupling_map
            metadata["backend_basis_gates"] = self.backend.basis_gates
            metadata["backend_num_qubits"] = self.backend.num_qubits

        if self.from_ham:
            result = self.energy
        else:
            result = self.phase

        result = Result(
            result=result,
            measurements=counts,
            parameters=None,
            metadata=metadata,
        )
        return result

    def set_backend(self, backend: QuantumBackend | None) -> None:
        """Attach a QuantumBackend instance for execution."""
        if backend is None:
            backend = TNQuantumBackend()
        self.backend = backend
        return



import copy
from timeit import default_timer
from typing import TypeAlias, Union

from numpy import ndarray
from qiskit import QuantumCircuit
from qiskit.circuit import CircuitInstruction, Operation
from sparse import SparseArray

from ..backend.base import QuantumBackend
from ..backend.tn_backend import TNQuantumBackend
from ..base import QuantumAlgorithm
from ..result import Result
from ..utils import add_controls, count_qubits, to_quantum_circuit

TypeOptions: TypeAlias = Union[
    QuantumCircuit, Operation, CircuitInstruction, ndarray, SparseArray  # type: ignore
]


class HadamardTest(QuantumAlgorithm):
    def __init__(
        self,
        unitary: TypeOptions,
        state: TypeOptions,
        backend: QuantumBackend | None = None,
    ) -> "HadamardTest":  # type: ignore
        """Constructor

        Args:
            unitary: The unitary operator to estimate the expectation value with respect to
            state: The state to estimate the expectation value with respect to
            backend: The backend to use, default to TN4QA circuit simulator
        """
        self.num_state_qubits = count_qubits(state)

        state_circ = to_quantum_circuit(state)
        unitary_circ = to_quantum_circuit(unitary)

        qc = QuantumCircuit(self.num_state_qubits + 1)
        qc.compose(
            unitary_circ, qubits=range(1, self.num_state_qubits + 1), inplace=True
        )
        qc = add_controls(qc, [0])
        qc.compose(
            state_circ,
            qubits=range(1, self.num_state_qubits + 1),
            inplace=True,
            front=True,
        )
        self._circuit = qc
        self.set_backend(backend=backend)

    @property
    def circuit(self) -> QuantumCircuit:
        return self._circuit

    def get_real_counts(self, num_shots: int) -> dict[str, int]:
        """Get counts for real part of expectation value"""
        qc_real = copy.deepcopy(self.circuit)
        qc_h = QuantumCircuit(self.num_state_qubits + 1)
        qc_h.h(0)
        qc_real.compose(qc_h, inplace=True, front=True)
        qc_real.h(0)
        real_counts = self.backend.run(qc_real, num_shots)
        return real_counts

    def get_imag_counts(self, num_shots: int) -> dict[str, int]:
        """Get counts for real part of expectation value"""
        qc_imag = copy.deepcopy(self.circuit)
        qc_prep = QuantumCircuit(self.num_state_qubits + 1)
        qc_prep.h(0)
        qc_prep.sdg(0)
        qc_imag.compose(qc_prep, inplace=True, front=True)
        qc_imag.h(0)
        imag_counts = self.backend.run(qc_imag, num_shots)
        return imag_counts

    def calculate_exp_value(
        self, real_counts: dict, imag_counts: dict, num_shots: int
    ) -> complex:
        """Calculate the expectation value given counts"""
        counts0_real = 0
        counts0_imag = 0
        for bitstring, count in real_counts.items():
            ancilla_bit = bitstring[0]
            if ancilla_bit == "0":
                counts0_real += count
        for bitstring, count in imag_counts.items():
            ancilla_bit = bitstring[0]
            if ancilla_bit == "0":
                counts0_imag += count
        real_part = 2 * (counts0_real / num_shots) - 1
        imag_part = 2 * (counts0_imag / num_shots) - 1
        return real_part + 1.0j * imag_part

    def run(self, num_shots: int = 1024) -> Result:
        """Run the full algorithm pipeline. Returns result object or final value.

        Args:
            num_shots: Number of shots to use per circuit
        """
        start_time = default_timer()
        real_counts = self.get_real_counts(num_shots)
        imag_counts = self.get_imag_counts(num_shots)
        self.expectation_value = self.calculate_exp_value(
            real_counts, imag_counts, num_shots
        )
        end_time = default_timer()

        metadata = {
            "algorithm_name": "Hadamard Test",
            "num_shots": num_shots,
            "total_runtime": end_time - start_time,
        }
        if self.backend is not None:
            metadata["backend_name"] = self.backend.name
            metadata["backend_coupling_map"] = self.backend.coupling_map
            metadata["backend_basis_gates"] = self.backend.basis_gates
            metadata["backend_num_qubits"] = self.backend.num_qubits
        result = Result(
            result=self.expectation_value,
            measurements=None,
            parameters=None,
            metadata=metadata,
        )
        return result

    def set_backend(self, backend: QuantumBackend | None) -> None:
        """Attach a QuantumBackend instance for execution."""
        if backend is None:
            backend = TNQuantumBackend()
        self.backend = backend
        return
